import logging
import traceback
from threading import Timer
from threading import Thread

from . import data_packets_socket
from .hpim_globals import SOURCE_LIFETIME
from .tree_interface import TreeInterface


class TreeInterfaceRootOriginator(TreeInterface):
    LOGGER = logging.getLogger('hpim.KernelEntry.RootInterface')

    def __init__(self, kernel_entry, interface_id, current_tree_state):
        extra_dict_logger = kernel_entry.kernel_entry_logger.extra.copy()
        extra_dict_logger['vif'] = interface_id
        extra_dict_logger['interfacename'] = kernel_entry.get_interface_name(interface_id)
        logger = logging.LoggerAdapter(TreeInterfaceRootOriginator.LOGGER, extra_dict_logger)
        TreeInterface.__init__(self, kernel_entry, interface_id, None, current_tree_state, logger)

        # Originator state
        self._source_active_timer = None
        self.set_source_active_timer()

        # TODO TESTE SOCKET RECV DATA PCKTS
        self.socket_is_enabled = True
        (s, g) = self.get_tree_id()
        interface_name = self.get_interface_name()
        self.socket_pkt = data_packets_socket.get_s_g_bpf_filter_code(s, g, interface_name)

        # run receive method in background
        receive_thread = Thread(target=self.socket_recv)
        receive_thread.daemon = True
        receive_thread.start()

        self.logger.debug('Created RootInterfaceOriginator')

    def socket_recv(self):
        """
        Socket used to receive data packets from the Root interface...
        Useful in order to control the tree state of Originator routers
        """
        while self.socket_is_enabled:
            try:
                self.socket_pkt.recvfrom(0)
                print("DATA RECEIVED")
                self.recv_data_msg()
            except:
                traceback.print_exc()
                continue
        print("EXIT SOCKET")

    ##########################################
    # Set timers
    ##########################################
    # Originator timers
    def set_source_active_timer(self):
        """
        Set Source Active timer
        """
        self.clear_source_active_timer()
        self._source_active_timer = Timer(SOURCE_LIFETIME, self.source_active_timeout)
        self._source_active_timer.start()

    def clear_source_active_timer(self):
        """
        Stop Source Active timer
        """
        if self._source_active_timer is not None:
            self._source_active_timer.cancel()


    ###########################################
    # Timer timeout
    ###########################################
    def source_active_timeout(self):
        """
        Source Active timer expired... react to this event
        """
        self._kernel_entry.sat_expires()

    ###########################################
    # Recv packets
    ###########################################
    def recv_data_msg(self):
        """
        Root interface received data packet
        """
        if not self.is_tree_inactive():
            self.set_source_active_timer()
            if self.is_tree_unsure():
                self._kernel_entry.sat_running()

    ###########################################
    # Change to in/out-tree
    ###########################################
    def node_is_out_tree(self):
        return

    def node_is_in_tree(self):
        return

    ####################################################################
    def is_forwarding(self):
        """
        This interface must not be included in the OIL of the multicast routing table, thus returning False
        """
        return False

    def delete(self):
        """
        Tree interface is being removed... due to change of interface roles or
        due to the removal of the tree by this router
        Clear all state from this interface regarding this tree
        """
        self.socket_is_enabled = False
        try:
            from socket import SHUT_RDWR
            self.socket_pkt.shutdown(SHUT_RDWR)
        except:
            pass
        self.socket_pkt.close()
        super().delete()
        self.clear_source_active_timer()
        self._source_active_timer = None



from threading import Timer, RLock
from hpimdm.tree.hpim_globals import MESSAGE_RETRANSMISSION_TIME, ACK_FAILURE_THRESHOLD


class ReliableMessageTransmission(object):
    def __init__(self, interface):
        self._interface = interface
        self._msg_multicast = None
        self._msg_unicast = {}
        self._neighbors_that_acked = set()
        self._number_of_failed_acks = {}
        self._retransmission_timer = None
        self._lock = RLock()

    def send_i_am_upstream(self, source, group, rpc):
        """
        Send reliably a new IamUpstream message
        """
        with self._lock:
            self.cancel_all_messages()

            (bt, sn) = self._interface.get_sequence_number()

            metric_preference = rpc.metric_preference
            metric = rpc.route_metric

            self._msg_multicast = self._interface.create_i_am_upstream_msg(my_boot_time=bt, sn=sn,
                                                                           source=source, group=group,
                                                                           metric_preference=metric_preference,
                                                                           metric=metric)

            self.set_retransmission_timer()
            self._interface.send(self._msg_multicast)

    def send_i_am_no_longer_upstream(self, source, group):
        """
        Send reliably a new IamNoLongerUpstream message
        """
        with self._lock:
            self.cancel_all_messages()

            (bt, sn) = self._interface.get_sequence_number()

            self._msg_multicast = self._interface.create_i_am_no_longer_upstream_msg(my_boot_time=bt, sn=sn,
                                                                                     source=source, group=group)

            self.set_retransmission_timer()
            self._interface.send(self._msg_multicast)

    def send_interest(self, source, group, dst):
        """
        Send reliably a new Interest message
        """
        with self._lock:
            self.cancel_message_unicast(dst)

            (bt, sn) = self._interface.get_sequence_number()

            packet = self._interface.create_interest_msg(my_boot_time=bt, sn=sn, source=source, group=group)
            self._msg_unicast[dst] = packet

            self.set_retransmission_timer()
            self._interface.send(packet, dst)

            # msg multicast doesnt require to be reliably protected to dst (because unicast msgs can only be transmitted
            # in concurrency with IamNoLongerUpstream - interest msg implies that neighbor is not Upstream... so dont
            # wait ack regarding multicast msg)
            if self._msg_multicast is not None:
                self._neighbors_that_acked.add(dst)

    def send_no_interest(self, source, group, dst):
        """
        Send reliably a new NoInterest message
        """
        with self._lock:
            self.cancel_message_unicast(dst)

            (bt, sn) = self._interface.get_sequence_number()

            packet = self._interface.create_no_interest_msg(my_boot_time=bt, sn=sn, source=source, group=group)
            self._msg_unicast[dst] = packet

            self.set_retransmission_timer()
            self._interface.send(packet, dst)

            # msg multicast doesnt require to be reliably protected to dst (because unicast msgs can only be transmitted
            # in concurrency with IamNoLongerUpstream - interest msg implies that neighbor is not Upstream... so dont
            # wait ack regarding multicast msg)
            if self._msg_multicast is not None:
                self._neighbors_that_acked.add(dst)

    def receive_ack(self, neighbor_ip, bt, sn):
        """
        Received Ack regarding this tree
        """
        with self._lock:
            msg = self._msg_multicast
            if msg is not None and (bt > msg.payload.boot_time or
                                    bt == msg.payload.boot_time and sn >= msg.payload.payload.sequence_number):
                self._neighbors_that_acked.add(neighbor_ip)
                if self.did_all_neighbors_acked():
                    self.cancel_messsage_multicast()

            if neighbor_ip in self._msg_unicast:
                msg = self._msg_unicast[neighbor_ip]
                if bt > msg.payload.boot_time or (bt == msg.payload.boot_time and
                                                  sn >= msg.payload.payload.sequence_number):
                    self.cancel_message_unicast(neighbor_ip)

    def did_all_neighbors_acked(self):
        """
        Verify if all known neighbors have acked a multicast message
        """
        with self._lock:
            return self._interface.did_all_neighbors_acked(self._neighbors_that_acked)

    def cancel_messsage_multicast(self):
        """
        Stop reliably monitoring an IamUpstream/IamNoLongerUpstream message
        """
        with self._lock:
            self._neighbors_that_acked.clear()
            self._msg_multicast = None
            if len(self._msg_unicast) == 0:
                # there are no multicast nor unicast messages being transmitted
                self.clear_retransmission_timer()
                self._number_of_failed_acks.clear()
            else:
                # only unicast messages are being transmitted
                neighbors_that_have_acked = self._number_of_failed_acks.keys() - self._msg_unicast.keys()
                for n in neighbors_that_have_acked:
                    self._number_of_failed_acks.pop(n, None)

    def cancel_message_unicast(self, ip):
        """
        Stop reliably monitoring an Interest/NoInterest message
        """
        with self._lock:
            self._msg_unicast.pop(ip, None)
            self._number_of_failed_acks.pop(ip, None)
            if self._msg_multicast is None and len(self._msg_unicast) == 0:
                self.clear_retransmission_timer()

    def cancel_all_messages(self):
        """
        Stop reliably monitoring any message regarding this tree
        (IamUpstream/IamNoLongerUpstream/Interest/NoInterest)
        """
        with self._lock:
            self.clear_retransmission_timer()
            self._neighbors_that_acked.clear()
            self._number_of_failed_acks.clear()
            self._msg_multicast = None
            self._msg_unicast.clear()

    ##########################################
    # Set timers
    ##########################################
    # Reliable timer
    def set_retransmission_timer(self):
        """
        Set retransmission timer used to control retransmission of control messages
        """
        self.clear_retransmission_timer()
        self._retransmission_timer = Timer(MESSAGE_RETRANSMISSION_TIME, self.retransmission_timeout)
        self._retransmission_timer.start()

    def clear_retransmission_timer(self):
        """
        Stop retransmission timer
        """
        if self._retransmission_timer is not None:
            self._retransmission_timer.cancel()

    ###########################################
    # Timer timeout
    ###########################################
    def retransmission_timeout(self):
        """
        Retransmission timer has expired
        """
        neighbors_not_acked = set()
        with self._interface.neighbors_lock:
            with self._lock:
                # recheck if all neighbors acked
                if self._msg_multicast is not None and self.did_all_neighbors_acked():
                    self.cancel_messsage_multicast()
                elif self._msg_multicast is not None:
                    # take note of all neighbors that have not acked the multicast msg
                    neighbors_not_acked = self.get_interface_neighbors() - self._neighbors_that_acked

                # didnt received acks from every neighbor... so lets resend msg and reschedule timer
                msg = self._msg_multicast
                if msg is not None:
                    self._interface.send(msg)

                for (dst, msg) in self._msg_unicast.copy().items():
                    if self._interface.is_neighbor(dst):
                        self._interface.send(msg, dst)
                        neighbors_not_acked.add(dst)  # take note of all neighbors that have not acked unicast messages
                    else:
                        self.cancel_message_unicast(dst)

                if self._msg_multicast is not None or len(self._msg_unicast) > 0:
                    self.set_retransmission_timer()

                # update number of failed acks per neighbor and check which ones should be considered to have failed
                for neighbor_ip in neighbors_not_acked:
                    self._number_of_failed_acks[neighbor_ip] = self._number_of_failed_acks.get(neighbor_ip, 0) + 1
                self.check_neighbor_failures()

    #############################################
    # Get Sequence Number for CheckpointSN
    #############################################
    def get_sequence_number(self):
        """
        Get the lowest sequence number of a control message that is being currently reliably transmitted...
        This method will be used to determine the CheckpointSN to be transmitted in Hello messages
        """
        bt_sn = (None, None)
        with self._lock:
            msg = self._msg_multicast
            if msg is not None:
                bt_sn = (msg.payload.boot_time, msg.payload.payload.sequence_number - 1)

            for (_, msg) in self._msg_unicast.items():
                if bt_sn[0] is None or bt_sn[0] is not None and \
                        (bt_sn[0] < msg.payload.boot_time or
                         bt_sn[0] == msg.payload.boot_time and bt_sn[1] > (msg.payload.payload.sequence_number - 1)):
                    bt_sn = (msg.payload.boot_time, msg.payload.payload.sequence_number - 1)
        return bt_sn

    #############################################
    # Check neighbor failures
    # Force neighbor failure in case neighbor does not ack successive control messages
    #############################################
    def check_neighbor_failures(self):
        with self._lock:
            for (neighbor_ip, ack_failures) in self._number_of_failed_acks.copy().items():
                if ack_failures > ACK_FAILURE_THRESHOLD:
                    print("NEIGHBOR FAILED DUE TO ACK LACK: " + neighbor_ip)
                    self.force_neighbor_failure(neighbor_ip)

    def force_neighbor_failure(self, neighbor_ip):
        with self._lock:
            self._interface.force_neighbor_failure(neighbor_ip)
            self._number_of_failed_acks.pop(neighbor_ip, None)

    def get_interface_neighbors(self):
        return self._interface.get_neighbors_ip()



from abc import ABCMeta, abstractmethod


class SFMRAssertABC(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def is_assert_winner():
        raise Exception


class SFMRAssertWinner(SFMRAssertABC):
    @staticmethod
    def is_assert_winner():
        """
        Determine if this state is AssertWinner
        """
        return True

    def __str__(self):
        return "AssertWinner"


class SFMRAssertLoser(SFMRAssertABC):
    @staticmethod
    def is_assert_winner():
        """
        Determine if this state is AssertWinner
        """
        return False

    def __str__(self):
        return "AssertLoser"


class AssertState():
    Winner = SFMRAssertWinner()
    Loser = SFMRAssertLoser()



import time
import logging
import ipaddress
from threading import Timer

from hpimdm.utils import TYPE_CHECKING
from hpimdm.tree.metric import AssertMetric
from hpimdm.tree import hpim_globals
from hpimdm.tree.hpim_globals import HELLO_PERIOD
from hpimdm.packet.Packet import Packet
if hpim_globals.MSG_FORMAT == "BINARY":
    from hpimdm.packet.PacketHPIMSync import PacketHPIMSync
    from hpimdm.packet.PacketHPIMHelloOptions import PacketHPIMHelloHoldtime
    from hpimdm.packet.PacketHPIMHeader import PacketHPIMHeader
else:
    from hpimdm.packet.PacketHPIMSync import PacketHPIMSyncJson as PacketHPIMSync
    from hpimdm.packet.PacketHPIMHelloOptions import PacketHPIMHelloHoldtimeJson as PacketHPIMHelloHoldtime
    from hpimdm.packet.PacketHPIMHeader import PacketHPIMHeaderJson as PacketHPIMHeader

if TYPE_CHECKING:
    from hpimdm.InterfaceHPIM import InterfaceHPIM

DEFAULT_HELLO_HOLD_TIME_DURING_SYNC = 4 * hpim_globals.SYNC_RETRANSMISSION_TIME
DEFAULT_HELLO_HOLD_TIME_AFTER_SYNC = 120


class NeighborState:
    @staticmethod
    def new_neighbor_or_adjacency_reset(neighbor):
        neighbor.set_sync_state(Slave)

        neighbor.start_snapshot()

        # Remove all info from neighbor (if already knew it)
        neighbor.tree_interest_state.clear()
        neighbor.tree_metric_state.clear()
        neighbor.last_sequence_number.clear()
        neighbor.current_sync_sn = 0
        neighbor.neighbor_snapshot_sn = 0
        neighbor.checkpoint_sn = 0
        #

        my_snapshot_mrt = neighbor.my_snapshot_multicast_routing_table[0:neighbor.sync_fragmentation]
        my_more_bit = len(neighbor.my_snapshot_multicast_routing_table) > 0
        my_snapshot_sn = neighbor.my_snapshot_sequencer

        pkt_s = PacketHPIMSync(my_snapshot_sn, 0,
                               sync_sn=neighbor.current_sync_sn, upstream_trees=my_snapshot_mrt,
                               master_flag=True, more_flag=my_more_bit,
                               neighbor_boot_time=neighbor.time_of_boot)
        pkt = Packet(payload=PacketHPIMHeader(pkt_s, neighbor.my_snapshot_boot_time))
        neighbor.contact_interface.send(pkt, neighbor.ip)
        neighbor.set_sync_timer()
        neighbor.set_hello_hold_time(DEFAULT_HELLO_HOLD_TIME_DURING_SYNC)

    @staticmethod
    def recv_sync(neighbor, tree_state, my_snapshot_sn, neighbor_snapshot_sn, sync_sn, master_bit, more_bit, hello_options):
        return

    @staticmethod
    def sync_timer_expires(neighbor):
        return


class Synced(NeighborState):
    @staticmethod
    def recv_sync(neighbor, tree_state, my_snapshot_sn, neighbor_snapshot_sn, sync_sn, master_bit, more_bit, hello_options):
        if neighbor.neighbor_snapshot_sn > neighbor_snapshot_sn:
            return
        elif neighbor.neighbor_snapshot_sn < neighbor_snapshot_sn:
            Synced.new_neighbor_or_adjacency_reset(neighbor)
            return

        if sync_sn != neighbor.current_sync_sn:
            return

        if master_bit and neighbor.my_snapshot_sequencer == my_snapshot_sn:
            pkt_s = PacketHPIMSync(my_snapshot_sn, neighbor_snapshot_sn, sync_sn=sync_sn,
                                   master_flag=False, more_flag=False,
                                   neighbor_boot_time=neighbor.time_of_boot)
            pkt_s.add_hello_option(PacketHPIMHelloHoldtime(holdtime=4 * HELLO_PERIOD))
            pkt = Packet(payload=PacketHPIMHeader(pkt_s, neighbor.my_snapshot_boot_time))
            neighbor.send(pkt)


class Master(NeighborState):
    @staticmethod
    def recv_sync(neighbor, tree_state, my_snapshot_sn, neighbor_snapshot_sn, sync_sn, master_bit, more_bit, hello_options):
        if neighbor.current_sync_sn == 0 and neighbor.neighbor_snapshot_sn == 0:
            neighbor.neighbor_snapshot_sn = neighbor_snapshot_sn

        if neighbor.neighbor_snapshot_sn > neighbor_snapshot_sn:
            return
        elif neighbor.neighbor_snapshot_sn < neighbor_snapshot_sn:
            Master.new_neighbor_or_adjacency_reset(neighbor)
            return

        if sync_sn != neighbor.current_sync_sn:
            return

        if master_bit and (sync_sn > 0 and neighbor.my_snapshot_sequencer == my_snapshot_sn or sync_sn == 0):
            neighbor.install_tree_state(tree_state)

            my_snapshot_mrt = neighbor.my_snapshot_multicast_routing_table[neighbor.current_sync_sn*neighbor.sync_fragmentation:
                                                                           (neighbor.current_sync_sn+1)*neighbor.sync_fragmentation]
            my_more_bit = len(neighbor.my_snapshot_multicast_routing_table) > neighbor.current_sync_sn*neighbor.sync_fragmentation
            my_snapshot_sn = neighbor.my_snapshot_sequencer
            neighbor_sn = neighbor.neighbor_snapshot_sn

            pkt_s = PacketHPIMSync(my_snapshot_sn, neighbor_sn, sync_sn=neighbor.current_sync_sn,
                                   upstream_trees=my_snapshot_mrt, master_flag=False, more_flag=my_more_bit,
                                   neighbor_boot_time=neighbor.time_of_boot)
            if not my_more_bit:
                pkt_s.add_hello_option(PacketHPIMHelloHoldtime(holdtime=4 * HELLO_PERIOD))
            pkt = Packet(payload=PacketHPIMHeader(pkt_s, neighbor.my_snapshot_boot_time))
            neighbor.send(pkt)

            if sync_sn > 0 and not more_bit and not my_more_bit:
                if "HOLDTIME" in hello_options:
                    neighbor.set_hello_hold_time(hello_options["HOLDTIME"].holdtime)
                else:
                    neighbor.set_hello_hold_time(DEFAULT_HELLO_HOLD_TIME_AFTER_SYNC)
                neighbor.set_sync_state(Synced)
                neighbor.clear_sync_timer()
                del neighbor.my_snapshot_multicast_routing_table[:]
            else:
                neighbor.set_hello_hold_time(DEFAULT_HELLO_HOLD_TIME_DURING_SYNC)
                neighbor.set_sync_timer()
                neighbor.current_sync_sn += 1

    @staticmethod
    def sync_timer_expires(neighbor):
        my_snapshot_mrt = neighbor.my_snapshot_multicast_routing_table[(neighbor.current_sync_sn - 1) * neighbor.sync_fragmentation:
                                                                       neighbor.current_sync_sn * neighbor.sync_fragmentation]
        my_more_bit = len(neighbor.my_snapshot_multicast_routing_table) > (neighbor.current_sync_sn - 1) * neighbor.sync_fragmentation
        my_snapshot_sn = neighbor.my_snapshot_sequencer
        neighbor_sn = neighbor.neighbor_snapshot_sn

        pkt_s = PacketHPIMSync(my_snapshot_sn, neighbor_sn, sync_sn=neighbor.current_sync_sn - 1,
                               upstream_trees=my_snapshot_mrt, master_flag=False, more_flag=my_more_bit,
                               neighbor_boot_time=neighbor.time_of_boot)
        if not my_more_bit:
            pkt_s.add_hello_option(PacketHPIMHelloHoldtime(holdtime=4 * HELLO_PERIOD))

        pkt = Packet(payload=PacketHPIMHeader(pkt_s, neighbor.my_snapshot_boot_time))
        neighbor.send(pkt)
        neighbor.set_sync_timer()


class Slave(NeighborState):
    @staticmethod
    def recv_sync(neighbor, tree_state, my_snapshot_sn, neighbor_snapshot_sn, sync_sn, master_bit, more_bit, hello_options):
        if neighbor.current_sync_sn == 0 and neighbor.neighbor_snapshot_sn == 0:
            neighbor.neighbor_snapshot_sn = neighbor_snapshot_sn

        if neighbor.neighbor_snapshot_sn > neighbor_snapshot_sn:
            return
        elif neighbor.neighbor_snapshot_sn < neighbor_snapshot_sn:
            Slave.new_neighbor_or_adjacency_reset(neighbor)
            return

        if sync_sn != neighbor.current_sync_sn:
            return

        if sync_sn == 0 and master_bit:
            my_ip = ipaddress.ip_address(neighbor.contact_interface.get_ip())
            neighbor_ip = ipaddress.ip_address(neighbor.ip)
            if my_ip < neighbor_ip:
                neighbor.set_sync_state(Master)
                neighbor.current_sync_sn = 0
                Master.recv_sync(neighbor, tree_state, my_snapshot_sn, neighbor_snapshot_sn, sync_sn, master_bit,
                                 more_bit, hello_options)
            else:
                Slave.sync_timer_expires(neighbor)
        elif not master_bit and neighbor.my_snapshot_sequencer == my_snapshot_sn:
            my_more_bit = len(neighbor.my_snapshot_multicast_routing_table) > neighbor.current_sync_sn*neighbor.sync_fragmentation

            if sync_sn > 0 and not my_more_bit and not more_bit:
                if "HOLDTIME" in hello_options:
                    neighbor.set_hello_hold_time(hello_options["HOLDTIME"].holdtime)
                else:
                    neighbor.set_hello_hold_time(DEFAULT_HELLO_HOLD_TIME_AFTER_SYNC)
                neighbor.set_sync_state(Synced)
                neighbor.clear_sync_timer()
                del neighbor.my_snapshot_multicast_routing_table[:]
            else:
                neighbor.set_hello_hold_time(DEFAULT_HELLO_HOLD_TIME_DURING_SYNC)
                neighbor.current_sync_sn += 1
                neighbor.install_tree_state(tree_state)

                my_snapshot_mrt = neighbor.my_snapshot_multicast_routing_table[neighbor.current_sync_sn*neighbor.sync_fragmentation:
                                                                               (neighbor.current_sync_sn+1)*neighbor.sync_fragmentation]
                my_more_bit = len(neighbor.my_snapshot_multicast_routing_table) > neighbor.current_sync_sn*neighbor.sync_fragmentation
                my_snapshot_sn = neighbor.my_snapshot_sequencer
                neighbor_sn = neighbor.neighbor_snapshot_sn

                pkt_s = PacketHPIMSync(my_snapshot_sn, neighbor_sn, sync_sn=neighbor.current_sync_sn,
                                       upstream_trees=my_snapshot_mrt, master_flag=True, more_flag=my_more_bit,
                                       neighbor_boot_time=neighbor.time_of_boot)
                if not my_more_bit:
                    pkt_s.add_hello_option(PacketHPIMHelloHoldtime(holdtime=4 * HELLO_PERIOD))
                pkt = Packet(payload=PacketHPIMHeader(pkt_s, neighbor.my_snapshot_boot_time))
                neighbor.send(pkt)
                neighbor.set_sync_timer()

    @staticmethod
    def sync_timer_expires(neighbor):
        my_snapshot_mrt = neighbor.my_snapshot_multicast_routing_table[neighbor.current_sync_sn * neighbor.sync_fragmentation:
                                                                       (neighbor.current_sync_sn + 1) * neighbor.sync_fragmentation]
        my_more_bit = len(neighbor.my_snapshot_multicast_routing_table) > neighbor.current_sync_sn * neighbor.sync_fragmentation
        my_snapshot_sn = neighbor.my_snapshot_sequencer
        neighbor_sn = neighbor.neighbor_snapshot_sn

        pkt_s = PacketHPIMSync(my_snapshot_sn, neighbor_sn, sync_sn=neighbor.current_sync_sn,
                               upstream_trees=my_snapshot_mrt, master_flag=True, more_flag=my_more_bit,
                               neighbor_boot_time=neighbor.time_of_boot)
        if not my_more_bit:
            pkt_s.add_hello_option(PacketHPIMHelloHoldtime(holdtime=4 * HELLO_PERIOD))

        pkt = Packet(payload=PacketHPIMHeader(pkt_s, neighbor.my_snapshot_boot_time))
        neighbor.send(pkt)
        neighbor.set_sync_timer()


class Unknown(NeighborState):
    @staticmethod
    def recv_sync(neighbor, tree_state, my_snapshot_sn, neighbor_snapshot_sn, sync_sn, master_bit, more_bit, hello_options):
        if sync_sn == 0 and sync_sn == neighbor.current_sync_sn and master_bit:
            neighbor.set_sync_state(Master)
            neighbor.set_hello_hold_time(DEFAULT_HELLO_HOLD_TIME_DURING_SYNC)

            neighbor.start_snapshot()

            # Remove all info from neighbor (if already knew it)
            neighbor.tree_interest_state.clear()
            neighbor.tree_metric_state.clear()
            neighbor.last_sequence_number.clear()
            neighbor.current_sync_sn = 0
            neighbor.neighbor_snapshot_sn = 0
            neighbor.checkpoint_sn = 0
            #

            Master.recv_sync(neighbor, tree_state, my_snapshot_sn, neighbor_snapshot_sn, sync_sn, master_bit, more_bit, hello_options)
        else:
            Unknown.new_neighbor_or_adjacency_reset(neighbor)


class Neighbor:
    LOGGER = logging.getLogger('hpim.Interface.Neighbor')

    def __init__(self, contact_interface: "InterfaceHPIM", ip, hello_hold_time: int, neighbor_time_of_boot: int,
                 my_interface_boot_time: int):
        if hello_hold_time == hpim_globals.HELLO_HOLD_TIME_TIMEOUT:
            raise Exception
        logger_info = dict(contact_interface.interface_logger.extra)
        logger_info['neighbor_ip'] = ip
        self.neighbor_logger = logging.LoggerAdapter(self.LOGGER, logger_info)

        self.contact_interface = contact_interface
        self.ip = ip
        self.time_of_boot = neighbor_time_of_boot
        self.neighbor_snapshot_sn = 0

        self.neighbor_liveness_timer = None
        self.hello_hold_time = None
        self.set_hello_hold_time(hello_hold_time)
        self.time_of_last_update = time.time()

        self.current_sync_sn = 0
        self.sync_fragmentation = 0

        # Tree Database storage
        self.tree_interest_state = {}
        self.tree_metric_state = {}

        # Control if received control packets should be processed
        # Used to detect msg retransmissions and out of order reception
        self.last_sequence_number = {}

        self.sync_timer = None
        self.neighbor_state = Unknown
        self.neighbor_logger.debug('Neighbor state of ' + self.ip + ' transitions to ' + self.neighbor_state.__name__)

        # checkpoint sn
        self.checkpoint_sn = 0

        # Information of my snapshot
        self.my_snapshot_boot_time = my_interface_boot_time
        self.my_snapshot_sequencer = 0
        self.my_snapshot_multicast_routing_table = []

    ######################################################################
    # Sync Timer
    ######################################################################
    def set_sync_timer(self):
        """
        Set Sync timer... useful when the Sync process is making progress and a Sync message from the neighbor node must be received
        """
        self.clear_sync_timer()
        self.sync_timer = Timer(hpim_globals.SYNC_RETRANSMISSION_TIME, self.sync_timeout)
        self.sync_timer.start()

    def clear_sync_timer(self):
        """
        Cancel Sync timer... useful when the Sync process finishes
        """
        if self.sync_timer is not None:
            self.sync_timer.cancel()

    ###########################################
    # Sync Timer timeout
    ###########################################
    def sync_timeout(self):
        """
        Expiration of Sync timer (mus cause a retransmission of a Sync message)
        """
        self.neighbor_state.sync_timer_expires(self)

    ######################################################################
    # Neighbor Liveness Timer
    ######################################################################
    def set_hello_hold_time(self, hello_hold_time: int):
        """
        Set Neighbor liveness timer due to progress in Sync process or a received Hello message
        """
        self.hello_hold_time = hello_hold_time
        if self.neighbor_liveness_timer is not None:
            self.neighbor_liveness_timer.cancel()

        if hello_hold_time == hpim_globals.HELLO_HOLD_TIME_TIMEOUT:
            self.remove()
        else:
            self.neighbor_liveness_timer = Timer(hello_hold_time, self.remove)
            self.neighbor_liveness_timer.start()

    ###########################################
    # Neighbor Liveness Timer timeout
    ###########################################
    def remove(self):
        """
        Remove neighbor node because neighbor liveness timer expired
        """
        with self.contact_interface.neighbors_lock:
            print('HELLO TIMER EXPIRED... remove neighbor')
            self.remove_neighbor_state()
            self.contact_interface.remove_neighbor(self.ip)

    ############################################
    # Sync State
    ############################################
    def set_sync_state(self, state):
        """
        Set sync state of this neighbor node (Unknown or Master or Slave or Synced)
        """
        if self.neighbor_state == state:
            return

        self.neighbor_state = state
        self.neighbor_logger.debug('Neighbor state of ' + self.ip + ' transitions to ' + state.__name__ +
                                   ' with MyBootTime=' + str(self.my_snapshot_boot_time) +
                                   '; MySnapshotSN=' + str(self.my_snapshot_sequencer) +
                                   '; NeighborBootTime=' + str(self.time_of_boot) +
                                   '; NeighborSnapshotSN=' + str(self.neighbor_snapshot_sn))
        if state == Synced:
            self.contact_interface.get_kernel().recheck_all_trees(self.contact_interface.vif_index)

    def install_tree_state(self, tree_state: list):
        """
        Store Upstream state regarding trees that were included in a Sync message... Since we allow installing state from
        IamUpstream/IamNoLongerUpstream/Interest/NoInterest messages concurrently to an ongoing synchronization, verify if
        trees included in Sync message have state fresher than the one that is already stored (in a non-Sync message)
        """
        for t in tree_state:
            tree_id = (t.source, t.group)
            if self.last_sequence_number.get(tree_id, 0) > self.neighbor_snapshot_sn:
                continue

            self.tree_metric_state[tree_id] = AssertMetric(metric_preference=t.metric_preference,
                                                           route_metric=t.metric, ip_address=self.ip)

    def remove_tree_state(self, source, group):
        """
        Remove all stored state of the neighbor node regarding trees in Unknown state
        """
        self.tree_interest_state.pop((source, group), None)

    def get_known_trees(self):
        """
        Get all trees that I am storing state regarding this neighbor node
        """
        a = set(self.tree_metric_state.keys())
        b = set(self.tree_interest_state.keys())
        return a.union(b)

    ######################################################################
    # Send Messages
    ######################################################################
    def send(self, packet):
        """
        Send messages destined to this neighbor node... Used in the neighbor state machine implementation
        """
        self.contact_interface.send(packet, self.ip)

    ######################################################################
    # Receive Messages
    ######################################################################
    def recv_hello(self, boot_time, holdtime, checkpoint_sn):
        """
        Process a received Hello message from this neighbor node
        """
        if boot_time < self.time_of_boot:
            return
        elif boot_time > self.time_of_boot:
            self.time_of_boot = boot_time
            self.neighbor_snapshot_sn = 0
            self.neighbor_state.new_neighbor_or_adjacency_reset(self)
            return

        if self.neighbor_state == Synced:
            self.time_of_last_update = time.time()
            self.set_hello_hold_time(holdtime)
            self.set_checkpoint_sn(checkpoint_sn)
        elif holdtime == 0:
            self.set_hello_hold_time(holdtime)

    def recv_sync(self, upstream_trees, my_sn, neighbor_sn, boot_time, sync_sn, master_flag, more_flag, own_interface_boot_time, hello_options):
        """
        Process a received Sync message from this neighbor node
        """
        if boot_time < self.time_of_boot:
            return
        elif boot_time > self.time_of_boot or own_interface_boot_time > self.my_snapshot_boot_time:
            self.time_of_boot = boot_time
            self.neighbor_snapshot_sn = 0
            self.neighbor_state.new_neighbor_or_adjacency_reset(self)
            return

        self.neighbor_state.recv_sync(self, upstream_trees, my_sn, neighbor_sn, sync_sn, master_flag, more_flag, hello_options)

    def recv_reliable_packet(self, sn, tree, boot_time):
        """
        Decide if a packet received from this neighbor should be processed
        """
        if boot_time < self.time_of_boot:
            return False
        elif boot_time > self.time_of_boot:
            self.time_of_boot = boot_time
            self.neighbor_snapshot_sn = 0
            self.start_sync_process()
            return False

        if self.neighbor_state == Unknown or self.current_sync_sn == 0:
            #do not interpret control message without having the guarantee of
            # correct <NeighborBootTime; NeighborSnapshotSN> pair
            return False

        last_received_sn = self.last_sequence_number.get(tree, 0)

        if sn <= self.neighbor_snapshot_sn or sn <= self.checkpoint_sn:
            # dont deliver to application
            print("RCVD ", sn)
            print("NSSN ", self.neighbor_snapshot_sn)
            return False
        elif sn >= last_received_sn:
            (source, group) = tree
            packet = self.contact_interface.create_ack_msg(my_boot_time=self.contact_interface.time_of_boot, sn=sn,
                                                           source=source, group=group,
                                                           neighbor_boot_time=boot_time,
                                                           neighbor_snapshot_sn=self.neighbor_snapshot_sn,
                                                           my_snapshot_sn=self.my_snapshot_sequencer)
            self.contact_interface.send(packet, self.ip)

            if sn > last_received_sn:
                # update most recent sn received from this neighbor
                self.last_sequence_number[tree] = sn

                # deliver to application
                return True
        # dont deliver to application
        print("RCVD ", sn)
        print("LAST TREE SN ", last_received_sn)
        return False

    def recv_ack(self, my_boot_time, neighbor_boot_time, my_snapshot_sn, neighbor_snapshot):
        """
        Decide if a received Ack should be processed... this decision is based on the SNs obtained during the Sync
        process with this neighbor
        """
        if neighbor_boot_time < self.time_of_boot:
            return False
        elif neighbor_boot_time > self.time_of_boot:
            self.time_of_boot = neighbor_boot_time
            self.neighbor_snapshot_sn = 0
            self.start_sync_process()
            return False

        return self.neighbor_state != Unknown and self.current_sync_sn > 0 and \
               self.my_snapshot_boot_time == my_boot_time and self.time_of_boot == neighbor_boot_time and\
               self.my_snapshot_sequencer == my_snapshot_sn and self.neighbor_snapshot_sn == neighbor_snapshot

    #####################################################
    # CheckpointSN... Store and clear lower SNs
    #####################################################
    def set_checkpoint_sn(self, checkpoint_sn):
        """
        By receiving an Hello message with a CheckpointSN store it (if greater than the previously stored CheckpointSN)...
        By storing a greater CheckpointSN, clear all SNs that are lower than the stored CheckpointSN
        """
        if checkpoint_sn > self.checkpoint_sn:
            self.checkpoint_sn = checkpoint_sn

            to_remove = {k for k, v in self.last_sequence_number.items() if v <= checkpoint_sn}
            for k in to_remove:
                self.last_sequence_number.pop(k)

    #######################################################
    # Synchronization methods for starting it
    #######################################################
    def start_sync_process(self):
        """
        Trigger synchronization with this neighbor node
        """
        self.neighbor_state.new_neighbor_or_adjacency_reset(self)

    def start_snapshot(self):
        """
        Create my own snapshot and set my SNs (my BootTime and MySnapshotSN)
        """
        (my_snapshot_bt, my_snapshot_sn, my_snapshot_mrt) = self.contact_interface.snapshot_multicast_routing_table()
        self.my_snapshot_boot_time = my_snapshot_bt
        self.my_snapshot_sequencer = my_snapshot_sn
        self.my_snapshot_multicast_routing_table = list(my_snapshot_mrt.values())
        self.sync_fragmentation = hpim_globals.SYNC_FRAGMENTATION_MSG
        if self.sync_fragmentation == 0:
            sync_entry_len = len(self.contact_interface.create_sync_entry_hdr("", "", 0, 0))
            self.sync_fragmentation = (self.contact_interface.get_mtu() - self.contact_interface.get_ip_header_length()
                                       - 8 - self.contact_interface.security_len - 16) // sync_entry_len
        self.contact_interface.neighbor_start_synchronization(self.ip, my_snapshot_bt, my_snapshot_sn)

    #################################################################
    # Obtain Upstream and Interest information regarding a neighbor
    #################################################################
    def get_tree_state(self, tree):
        """
        Obtain Upstream and Interest state regarding neighbor node... This information is obtained based on previous
        messages received from this neighbor node that were stored in the neighbor structure
        """
        if self.neighbor_state != Synced:
            # do not interpret stored state if not Synced
            return (False, None)
        else:
            upstream_state = self.tree_metric_state.get(tree, None)
            interest_state = False
            if upstream_state is None:
                interest_state = self.tree_interest_state.get(tree, hpim_globals.INITIAL_FLOOD_ENABLED)
            print("INTEREST NEIGHBOR ", self.ip, ":", interest_state)
            print("UPSTREAM NEIGHBOR ", self.ip, ":", upstream_state)
            return (interest_state, upstream_state)

    #######################################
    # Remove state regarding neighbor
    #######################################
    def remove_neighbor_state(self):
        """
        Clear all information regarding neighbor node
        """
        self.neighbor_logger.debug('Removing neighbor ' + self.ip)
        if self.neighbor_liveness_timer is not None:
            self.neighbor_liveness_timer.cancel()

        self.clear_sync_timer()

        self.tree_interest_state.clear()
        self.tree_metric_state.clear()
        self.last_sequence_number.clear()
        del self.my_snapshot_multicast_routing_table[:]



# Allow flood initial data packet even when there is no information regarding
# Upstream neighbors
INITIAL_FLOOD_ENABLED = True
INITIAL_FLOOD_TIME = 15

# Define control packets format. Two formats are available: JSON and BINARY
MSG_FORMAT = "BINARY"

# Periodicity of Hello messages
HELLO_PERIOD = 10

# Originator will stop considering the tree Active after not hearing data packets
# from the source after SOURCE_LIFETIME SECONDS
SOURCE_LIFETIME = 210

# Periodicity for message retransmission
MESSAGE_RETRANSMISSION_TIME = 3

# Value to be transmitted on a Hello message after an interface is explicitly removed from the protocol.
# DONT CHANGE THIS VALUE!
HELLO_HOLD_TIME_TIMEOUT = 0

# Control fragmentation of Sync messages
# Number of trees per Sync message
# If zero use information from MTU of interface, otherwise only include a positive given number of trees per Sync message
SYNC_FRAGMENTATION_MSG = 0


# Number of ACKs that must be missed in order to declare a neighbor to have failed
# Use a number HIGHER than 1!!
ACK_FAILURE_THRESHOLD = 3


# Time to wait for a Sync message
# If Synchronization does not make progress, retransmit previous Sync message
SYNC_RETRANSMISSION_TIME = 3


# When an AW loses an assert and becomes AL, it can hold its forwarding state for a given amount of time to prevent
# the loss of data packets (the new AW may not be receiving multicast tree from its parent yet causing the interest
# signaling to be propagated upwards the tree).
# If enabled this prevents loss of data packets after the AW is replaced, but duplications may occur for a small
#   amount of time (the AW and the AL may both forwarding packets at the same time during AL_HOLD_FORWARDING_STATE_TIME)
# If disabled the duplication of data packets does not occur but loss of packets may occur
AL_HOLD_FORWARDING_STATE_ENABLED = True
AL_HOLD_FORWARDING_STATE_TIME = 2


# MULTIPLE TABLES SUPPORT
# Define which unicast routing table to be used for RPF checks and to get route metric information
# Default unicast routing table is 254
UNICAST_TABLE_ID = 254
# Define which multicast routing table to be used for setting multicast trees
# Default multicast routing table is 0
MULTICAST_TABLE_ID = 0



import logging
import traceback
from threading import Thread

from . import data_packets_socket
from .tree_interface import TreeInterface
from .root_state_machine import SFMRNewRootState #SFMRRootState


class TreeInterfaceRoot(TreeInterface):
    LOGGER = logging.getLogger('hpim.KernelEntry.RootInterface')

    def __init__(self, kernel_entry, interface_id, best_upstream_router, was_non_root, previous_tree_state, current_tree_state):
        extra_dict_logger = kernel_entry.kernel_entry_logger.extra.copy()
        extra_dict_logger['vif'] = interface_id
        extra_dict_logger['interfacename'] = kernel_entry.get_interface_name(interface_id)
        logger = logging.LoggerAdapter(TreeInterfaceRoot.LOGGER, extra_dict_logger)
        TreeInterface.__init__(self, kernel_entry, interface_id, best_upstream_router, current_tree_state, logger)

        # event 1
        if was_non_root and previous_tree_state.is_active() and current_tree_state.is_active():
            SFMRNewRootState.interfaces_roles_change_and_tree_remains_active(self)
        # event 2
        elif was_non_root and previous_tree_state.is_unsure() and current_tree_state.is_active():
            SFMRNewRootState.interfaces_roles_change_and_tree_was_unsure_and_transitions_to_active(self)
        # event 3
        elif was_non_root and previous_tree_state.is_active() and current_tree_state.is_unsure() and \
                best_upstream_router is None:
            SFMRNewRootState.interfaces_roles_change_and_tree_was_active_and_transitions_to_unsure_and_best_upstream_neighbor_is_null(self)
        # event 4
        elif was_non_root and previous_tree_state.is_active() and current_tree_state.is_unsure() and \
                best_upstream_router is not None:
            SFMRNewRootState.interfaces_roles_change_and_tree_was_active_and_transitions_to_unsure_and_best_upstream_neighbor_not_null(self)
        # event 5
        elif was_non_root and previous_tree_state.is_unsure() and current_tree_state.is_unsure() and \
                best_upstream_router is not None:
            SFMRNewRootState.interfaces_roles_change_and_tree_remains_unsure_and_best_upstream_neighbor_not_null(self)
        # event 6
        elif not was_non_root and best_upstream_router is not None:
            SFMRNewRootState.interfaces_roles_dont_change_and_best_upstream_neighbor_reelected(self)

        # Originator state
        # TODO TESTE SOCKET RECV DATA PCKTS
        self.socket_is_enabled = True
        (s, g) = self.get_tree_id()
        interface_name = self.get_interface_name()
        self.socket_pkt = data_packets_socket.get_s_g_bpf_filter_code(s, g, interface_name)

        # run receive method in background
        receive_thread = Thread(target=self.socket_recv)
        receive_thread.daemon = True
        receive_thread.start()

        self.logger.debug('Created RootInterface')


    def socket_recv(self):
        while self.socket_is_enabled:
            try:
                self.socket_pkt.recvfrom(0)
                print("DATA RECEIVED")
                self.recv_data_msg()
            except:
                traceback.print_exc()
                continue

    ###########################################
    # Recv packets
    ###########################################
    def recv_data_msg(self):
        """
        This root interface received a data packet
        """
        return

    def change_best_upstream_neighbor_state(self, best_upstream_neighbor_state):
        """
        A neighbor changed Upstream state due to the reception of any control packet
        (IamUpstream or IamNoLongerUpstream or Interest or NoInterest or Sync)
        """
        previous_best_upstream_router = self._best_upstream_router
        super().change_best_upstream_neighbor_state(best_upstream_neighbor_state)
        print(self.get_tree_id())
        print("UPSTREAM CHANGE ASSERT STATE")
        print("best:", previous_best_upstream_router)
        print("new best", best_upstream_neighbor_state)

        if best_upstream_neighbor_state is None:
            print("ASSERT IS NONE")
            return
        elif previous_best_upstream_router is None or previous_best_upstream_router is not best_upstream_neighbor_state:
            print("BEST UPSTREAM REELECTED")
            # EVENT 6 and 8
            SFMRNewRootState.interfaces_roles_dont_change_and_best_upstream_neighbor_reelected(self)

    ###########################################
    # Change to in/out-tree
    ###########################################
    def send_my_interest(self):
        """
        Send an Interest/NoInterest message through this interface based on the IT/OT state of this router
        """
        if self.is_node_in_tree():
            self.send_interest()
        else:
            self.send_no_interest()

    def node_is_out_tree(self):
        """
        Node is no longer interested in receiving data packets...
        React to this event in order to transmit some control packets
        """
        if self._best_upstream_router is not None:
            # event 7
            SFMRNewRootState.interfaces_roles_dont_change_and_router_transition_to_it_or_ot(self)

    def node_is_in_tree(self):
        """
        Node is no longer interested in receiving data packets...
        React to this event in order to transmit some control packets
        """
        if self._best_upstream_router is not None:
            # event 7
            SFMRNewRootState.interfaces_roles_dont_change_and_router_transition_to_it_or_ot(self)

    ####################################################################
    def is_forwarding(self):
        """
        This interface must not be included in the OIL of the multicast routing table, thus returning False
        """
        return False

    def delete(self):
        """
        Tree interface is being removed... due to change of interface roles or
        due to the removal of the tree by this router
        Clear all state from this interface regarding this tree
        """
        self.socket_is_enabled = False
        self.socket_pkt.close()
        super().delete()



import logging
from threading import Lock
from threading import Timer

from hpimdm.utils import TYPE_CHECKING
from .wrapper import NoListenersPresent
from .mld_globals import MULTICAST_LISTENER_INTERVAL, LAST_LISTENER_QUERY_INTERVAL

if TYPE_CHECKING:
    from .RouterState import RouterState


class GroupState(object):
    LOGGER = logging.getLogger('hpim.mld.RouterState.GroupState')

    def __init__(self, router_state: 'RouterState', group_ip: str):
        #logger
        extra_dict_logger = router_state.router_state_logger.extra.copy()
        extra_dict_logger['tree'] = '(*,' + group_ip + ')'
        self.group_state_logger = logging.LoggerAdapter(GroupState.LOGGER, extra_dict_logger)

        #timers and state
        self.router_state = router_state
        self.group_ip = group_ip
        self.state = NoListenersPresent
        self.timer = None
        self.retransmit_timer = None
        # lock
        self.lock = Lock()

        # KernelEntry's instances to notify change of igmp state
        self.multicast_interface_state = []
        self.multicast_interface_state_lock = Lock()

    def print_state(self):
        return self.state.print_state()

    ###########################################
    # Set state
    ###########################################
    def set_state(self, state):
        """
        Set membership state regarding this Group
        """
        self.state = state
        self.group_state_logger.debug("change membership state to: " + state.print_state())

    ###########################################
    # Set timers
    ###########################################
    def set_timer(self, alternative: bool = False, max_response_time: int = None):
        """
        Set timer
        """
        self.clear_timer()
        if not alternative:
            time = MULTICAST_LISTENER_INTERVAL
        else:
            time = self.router_state.interface_state.get_group_membership_time(max_response_time)

        timer = Timer(time, self.group_membership_timeout)
        timer.start()
        self.timer = timer

    def clear_timer(self):
        """
        Stop timer
        """
        if self.timer is not None:
            self.timer.cancel()

    def set_retransmit_timer(self):
        """
        Set retransmit timer
        """
        self.clear_retransmit_timer()
        retransmit_timer = Timer(LAST_LISTENER_QUERY_INTERVAL, self.retransmit_timeout)
        retransmit_timer.start()
        self.retransmit_timer = retransmit_timer

    def clear_retransmit_timer(self):
        """
        Stop retransmit timer
        """
        if self.retransmit_timer is not None:
            self.retransmit_timer.cancel()

    ###########################################
    # Get group state from specific interface state
    ###########################################
    def get_interface_group_state(self):
        """
        Get specific implementation of the membership state machine (depending on the querier state machine)
        """
        return self.state.get_state(self.router_state)

    ###########################################
    # Timer timeout
    ###########################################
    def group_membership_timeout(self):
        """
        Timer has expired
        """
        with self.lock:
            self.get_interface_group_state().group_membership_timeout(self)

    def retransmit_timeout(self):
        """
        Retransmit timer has expired
        """
        with self.lock:
            self.get_interface_group_state().retransmit_timeout(self)

    ###########################################
    # Receive Packets
    ###########################################
    def receive_report(self):
        """
        Received MLD Report packet regarding this group
        """
        with self.lock:
            self.get_interface_group_state().receive_report(self)

    def receive_done(self):
        """
        Received MLD Done packet regarding this group
        """
        with self.lock:
            self.get_interface_group_state().receive_done(self)

    def receive_group_specific_query(self, max_response_time: int):
        """
        Received MLD Group Specific Query packet regarding this group
        """
        with self.lock:
            self.get_interface_group_state().receive_group_specific_query(self, max_response_time)

    ###########################################
    # Notify Routing
    ###########################################
    def notify_routing_add(self):
        """
        Notify all tree entries that MLD considers to have hosts interested in this group
        """
        with self.multicast_interface_state_lock:
            print("notify+", self.multicast_interface_state)
            for interface_state in self.multicast_interface_state:
                interface_state.notify_membership(has_members=True)

    def notify_routing_remove(self):
        """
        Notify all tree entries that MLD considers to have not hosts interested in this group
        """
        with self.multicast_interface_state_lock:
            print("notify-", self.multicast_interface_state)
            for interface_state in self.multicast_interface_state:
                interface_state.notify_membership(has_members=False)

    def add_multicast_routing_entry(self, kernel_entry):
        """
        A new tree is being monitored by the multicast routing protocol that has the same group
        MLD will store these entries in order to accelerate the notification process regarding changes in MLD state
        """
        with self.multicast_interface_state_lock:
            self.multicast_interface_state.append(kernel_entry)
            return self.has_members()

    def remove_multicast_routing_entry(self, kernel_entry):
        """
        A tree is no longer being monitored by the multicast routing protocol
        Remove this tree from this object
        """
        with self.multicast_interface_state_lock:
            self.multicast_interface_state.remove(kernel_entry)

    def has_members(self):
        """
        Determine if MLD considers to have hosts interested in receiving data packets
        """
        return self.state is not NoListenersPresent

    def remove(self):
        """
        Remove this group from the MLD process
        Notify all trees that this group no longer considers to be connected to hosts
        This method will be invoked whenever an MLD interface is removed
        """
        with self.multicast_interface_state_lock:
            self.clear_retransmit_timer()
            self.clear_timer()
            for interface_state in self.multicast_interface_state:
                interface_state.notify_membership(has_members=False)
            del self.multicast_interface_state[:]



import struct
import socket

###########################################################################################################
# JSON FORMAT
###########################################################################################################
class PacketHPIMAckJson:
    PIM_TYPE = "ACK"

    def __init__(self, source, group, sequence_number, neighbor_boot_time=0, neighbor_snapshot_sn=0, my_snapshot_sn=0):
        self.source = source
        self.group = group
        self.neighbor_boot_time = neighbor_boot_time
        self.neighbor_snapshot_sn = neighbor_snapshot_sn
        self.my_snapshot_sn = my_snapshot_sn
        self.sequence_number = sequence_number

    def bytes(self) -> bytes:
        """
        Obtain Packet Ack in a format to be transmitted (JSON)
        """
        msg = {"SOURCE": self.source,
               "GROUP": self.group,
               "NEIGHBOR_BOOT_TIME": self.neighbor_boot_time,
               "NEIGHBOR_SNAPSHOT_SN": self.neighbor_snapshot_sn,
               "MY_SNAPSHOT_SN": self.my_snapshot_sn,
               "SN": self.sequence_number
              }

        return msg

    def __len__(self):
        return len(self.bytes())

    @classmethod
    def parse_bytes(cls, data: bytes):
        """
        Parse received Packet from JSON and create ProtocolAck object
        """
        source = data["SOURCE"]
        group = data["GROUP"]
        sn = data["SN"]
        nbt = data["NEIGHBOR_BOOT_TIME"]
        nssn = data["NEIGHBOR_SNAPSHOT_SN"]
        mssn = data["MY_SNAPSHOT_SN"]
        return cls(source, group, sn, nbt, nssn, mssn)


###########################################################################################################
# BINARY FORMAT
###########################################################################################################
'''
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                        Tree Source IP                         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Tree Group IP                         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                       Neighbor BootTime                       |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                       NeighborSnapshotSN                      |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                          MySnapshotSN                         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                     Neighbor Sequence Number                  |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketHPIMAck:
    PIM_TYPE = 6

    PIM_HDR_ACK = "! 4s 4s L L L L"
    PIM_HDR_ACK_LEN = struct.calcsize(PIM_HDR_ACK)
    FAMILY = socket.AF_INET

    def __init__(self, source_ip, group_ip, sequence_number, neighbor_boot_time=0, neighbor_snapshot_sn=0, my_snapshot_sn=0):
        if type(source_ip) not in (str, bytes) or type(group_ip) not in (str, bytes):
            raise Exception
        if type(source_ip) is bytes:
            source_ip = socket.inet_ntop(self.FAMILY, source_ip)
        if type(group_ip) is bytes:
            group_ip = socket.inet_ntop(self.FAMILY, group_ip)

        self.source = source_ip
        self.group = group_ip
        self.neighbor_boot_time = neighbor_boot_time
        self.sequence_number = sequence_number
        self.neighbor_snapshot_sn = neighbor_snapshot_sn
        self.my_snapshot_sn = my_snapshot_sn

    def bytes(self) -> bytes:
        """
        Obtain Packet Ack in a format to be transmitted (binary)
        """
        msg = struct.pack(self.PIM_HDR_ACK, socket.inet_pton(self.FAMILY, self.source),
                          socket.inet_pton(self.FAMILY, self.group), self.neighbor_boot_time, self.neighbor_snapshot_sn,
                          self.my_snapshot_sn, self.sequence_number)

        return msg

    def __len__(self):
        return len(self.bytes())

    @classmethod
    def parse_bytes(cls, data: bytes):
        """
        Parse received Packet from bits/bytes and convert them into ProtocolAck object
        """
        (tree_source, tree_group, neighbor_boot_time, neighbor_snapshot_sn, my_snapshot_sn, sn) =\
            struct.unpack(cls.PIM_HDR_ACK, data[:cls.PIM_HDR_ACK_LEN])
        return cls(tree_source, tree_group, sn, neighbor_boot_time, neighbor_snapshot_sn, my_snapshot_sn)


class PacketHPIMAck_v6(PacketHPIMAck):
    PIM_HDR_ACK = "! 16s 16s L L L L"
    PIM_HDR_ACK_LEN = struct.calcsize(PIM_HDR_ACK)
    FAMILY = socket.AF_INET6

    def __init__(self, source_ip, group_ip, sequence_number, neighbor_boot_time=0, neighbor_snapshot_sn=0, my_snapshot_sn=0):
        super().__init__(source_ip, group_ip, sequence_number, neighbor_boot_time, neighbor_snapshot_sn, my_snapshot_sn)



import struct
from abc import ABCMeta, abstractmethod
import math

###################################################################################
# JSON FORMAT
###################################################################################
class PacketHPIMHelloOptionsJson(metaclass=ABCMeta):
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |              Type             |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, hello_type: str):
        self.type = hello_type

    @abstractmethod
    def bytes(self) -> bytes:
        """
        Obtain Protocol Hello Option in a format to be transmitted (JSON)
        This method will return the Hello Option in JSON format
        """
        pass

    def __len__(self):
        # not used
        return 0

    @staticmethod
    def parse_bytes(data: tuple, hello_type: int = None):
        """
        Parse received Hello Option from JSON and convert it into Hello object
        """
        hello_type = data[0]
        data = data[1]
        return JSON_MSG_TYPES.get(hello_type, PacketHPIMHelloUnknownJson).parse_bytes(data, hello_type)


class PacketHPIMHelloHoldtimeJson(PacketHPIMHelloOptionsJson):
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |            Hold Time          |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''

    def __init__(self, holdtime: int or float):
        super().__init__(hello_type="HOLDTIME")
        self.holdtime = int(holdtime)

    def bytes(self) -> dict:
        """
        Obtain Protocol Hello HoldTime Option in a format to be transmitted (JSON)
        This method will return the Hello Option in JSON format
        """
        return {"HOLDTIME": self.holdtime}

    @staticmethod
    def parse_bytes(data, hello_type: int = None):
        """
        Parse received Hello Option HoldTime from JSON and convert it into Hello object
        """
        if hello_type is None:
            raise Exception
        holdtime = data
        return PacketHPIMHelloHoldtimeJson(holdtime=holdtime)



class PacketHPIMHelloCheckpointSNJson(PacketHPIMHelloOptionsJson):
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                         Checkpoint SN                         |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, checkpoint_sn: int):
        super().__init__(hello_type="CHECKPOINT_SN")
        self.checkpoint_sn = checkpoint_sn

    def bytes(self) -> dict:
        """
        Obtain Protocol Hello Option CheckpointSN in a format to be transmitted (JSON)
        This method will return the Hello Option in JSON format
        """
        return {"CHECKPOINT_SN": self.checkpoint_sn}

    @staticmethod
    def parse_bytes(data, hello_type: int = None):
        """
        Parse received Hello Option CheckpointSN from JSON and convert it into Hello object
        """
        if hello_type is None:
            raise Exception
        checkpoint_sn = data
        return PacketHPIMHelloCheckpointSNJson(checkpoint_sn=checkpoint_sn)


class PacketHPIMHelloUnknownJson(PacketHPIMHelloOptionsJson):
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                            Unknown                            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, hello_type):
        super().__init__(hello_type=hello_type)

    def bytes(self) -> bytes:
        """
        Unknown options are not transmitted... Throw exception
        """
        raise Exception

    @staticmethod
    def parse_bytes(data, hello_type: int = None):
        """
        In case the received Hello Option is unknown parse its content (get to know the length of this option
        in order to parse following options)
        """
        if hello_type is None:
            raise Exception
        return PacketHPIMHelloUnknownJson(hello_type)


JSON_MSG_TYPES = {"HOLDTIME": PacketHPIMHelloHoldtimeJson,
                  "CHECKPOINT_SN": PacketHPIMHelloCheckpointSNJson,
                 }



class PacketHPIMHelloOptions(metaclass=ABCMeta):
    TYPE = "UNKNOWN"
    PIM_HDR_OPTS = "! HH"
    PIM_HDR_OPTS_LEN = struct.calcsize(PIM_HDR_OPTS)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |              Type             |             Length            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, hello_type: int, length: int):
        self.type = hello_type
        self.length = length

    def bytes(self) -> bytes:
        """
        Obtain Protocol Hello Option in a format to be transmitted (binary)
        This method will return the Hello Option in binary format
        """
        return struct.pack(PacketHPIMHelloOptions.PIM_HDR_OPTS, self.type, self.length)

    def __len__(self):
        return self.PIM_HDR_OPTS_LEN + self.length

    @staticmethod
    def parse_bytes(data: bytes, hello_type: int = None, length: int = None):
        """
        Parse received Hello Option from binary and convert it into Hello object
        """
        (hello_type, length) = struct.unpack(PacketHPIMHelloOptions.PIM_HDR_OPTS,
                                             data[:PacketHPIMHelloOptions.PIM_HDR_OPTS_LEN])
        #print("hello_type:", type)
        #print("LENGTH:", length)
        data = data[PacketHPIMHelloOptions.PIM_HDR_OPTS_LEN:]
        #return PIM_MSG_TYPES[type](data)
        return NEW_PROTOCOL_MSG_TYPES.get(hello_type, PacketHPIMHelloUnknown).parse_bytes(data, hello_type, length)


class PacketHPIMHelloHoldtime(PacketHPIMHelloOptions):
    TYPE = "HOLDTIME"
    PIM_HDR_OPT = "! H"
    PIM_HDR_OPT_LEN = struct.calcsize(PIM_HDR_OPT)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |            Hold Time          |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, holdtime: int or float):
        super().__init__(hello_type=1, length=2)
        self.holdtime = int(holdtime)

    def bytes(self) -> bytes:
        """
        Obtain Protocol Hello HoldTime Option in a format to be transmitted (binary)
        This method will return the Hello Option in binary format
        """
        return super().bytes() + struct.pack(self.PIM_HDR_OPT, self.holdtime)

    @staticmethod
    def parse_bytes(data: bytes, hello_type: int = None, length: int = None):
        """
        Parse received Hello Option HoldTime from binary and convert it into Hello object
        """
        if hello_type is None or length is None:
            raise Exception
        (holdtime, ) = struct.unpack(PacketHPIMHelloHoldtime.PIM_HDR_OPT, data[:length])
        print("HOLDTIME:", holdtime)
        return PacketHPIMHelloHoldtime(holdtime=holdtime)


class PacketHPIMHelloCheckpointSN(PacketHPIMHelloOptions):
    TYPE = "CHECKPOINT_SN"
    PIM_HDR_OPT = "! L"
    PIM_HDR_OPT_LEN = struct.calcsize(PIM_HDR_OPT)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                         Checkpoint SN                         |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, checkpoint_sn: int):
        super().__init__(hello_type=2, length=4)
        self.checkpoint_sn = checkpoint_sn

    def bytes(self) -> bytes:
        """
        Obtain Protocol Hello CheckpointSN Option in a format to be transmitted (binary)
        This method will return the Hello Option in binary format
        """
        return super().bytes() + struct.pack(self.PIM_HDR_OPT, self.checkpoint_sn)

    @staticmethod
    def parse_bytes(data: bytes, hello_type: int = None, length: int = None):
        """
        Parse received Hello Option ChekpointSN from binary and convert it into Hello object
        """
        if hello_type is None or length is None:
            raise Exception
        (checkpoint_sn, ) = struct.unpack(PacketHPIMHelloCheckpointSN.PIM_HDR_OPT, data[:length])
        print("CheckpointSN:", checkpoint_sn)
        return PacketHPIMHelloCheckpointSN(checkpoint_sn=checkpoint_sn)


class PacketHPIMHelloUnknown(PacketHPIMHelloOptions):
    TYPE = "UNKNOWN"
    PIM_HDR_OPT = "! L"
    PIM_HDR_OPT_LEN = struct.calcsize(PIM_HDR_OPT)
    '''
     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                            Unknown                            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    '''
    def __init__(self, hello_type, length):
        super().__init__(hello_type=hello_type, length=length)
        #print("PIM Hello Option Unknown... TYPE=", type, "LENGTH=", length)

    def bytes(self) -> bytes:
        """
        Unknown options are not transmitted... Throw exception
        """
        raise Exception

    @staticmethod
    def parse_bytes(data: bytes, hello_type: int = None, length: int = None):
        """
        In case the received Hello Option is unknown parse its content (get to know the length of this option
        in order to parse following options)
        """
        if hello_type is None or length is None:
            raise Exception
        return PacketHPIMHelloUnknown(hello_type, length)


NEW_PROTOCOL_MSG_TYPES = {1: PacketHPIMHelloHoldtime,
                          2: PacketHPIMHelloCheckpointSN,
                         }



import random
import traceback
from threading import Timer, RLock
import socket
import time
import hmac
import hashlib
import logging
import netifaces

from hpimdm import Main
from hpimdm.Neighbor import Neighbor
from hpimdm.Interface import Interface
from hpimdm.tree.hpim_globals import MSG_FORMAT, HELLO_HOLD_TIME_TIMEOUT, HELLO_PERIOD
from hpimdm.ReliableMsgTransmission import ReliableMessageTransmission

from hpimdm.packet.Packet import Packet
from hpimdm.packet.ReceivedPacket import ReceivedPacket
if MSG_FORMAT == "BINARY":
    from hpimdm.packet.PacketHPIMHelloOptions import PacketHPIMHelloHoldtime, \
        PacketHPIMHelloCheckpointSN
    from hpimdm.packet.PacketHPIMHello import PacketHPIMHello
    from hpimdm.packet.PacketHPIMHeader import PacketHPIMHeader
    from hpimdm.packet.PacketHPIMSync import PacketHPIMSync, PacketHPIMSyncEntry
    from hpimdm.packet.PacketHPIMInterest import PacketHPIMNoInterest, PacketHPIMInterest
    from hpimdm.packet.PacketHPIMIamUpstream import PacketHPIMUpstream
    from hpimdm.packet.PacketHPIMNotUpstream import PacketHPIMNoLongerUpstream
    from hpimdm.packet.PacketHPIMAck import PacketHPIMAck
else:
    from hpimdm.packet.PacketHPIMHelloOptions import PacketHPIMHelloHoldtimeJson as PacketHPIMHelloHoldtime,\
        PacketHPIMHelloCheckpointSNJson as PacketHPIMHelloCheckpointSN
    from hpimdm.packet.PacketHPIMHello import PacketHPIMHelloJson as PacketHPIMHello
    from hpimdm.packet.PacketHPIMHeader import PacketHPIMHeaderJson as PacketHPIMHeader
    from hpimdm.packet.PacketHPIMSync import PacketHPIMSyncJson as PacketHPIMSync
    from hpimdm.packet.PacketHPIMIamUpstream import PacketHPIMUpstreamJson as PacketHPIMUpstream
    from hpimdm.packet.PacketHPIMInterest import PacketHPIMNoInterestJson as PacketHPIMNoInterest,\
        PacketHPIMInterestJson as PacketHPIMInterest
    from hpimdm.packet.PacketHPIMSync import PacketHPIMSyncEntryJson as PacketHPIMSyncEntry
    from hpimdm.packet.PacketHPIMNotUpstream import PacketHPIMNoLongerUpstreamJson as PacketHPIMNoLongerUpstream
    from hpimdm.packet.PacketHPIMAck import PacketHPIMAckJson as PacketHPIMAck


class InterfaceHPIM(Interface):
    MCAST_GRP = '224.0.0.13'

    MAX_SEQUENCE_NUMBER = (2**32-1)#45 <- test with lower MAXIMUM_SEQUENCE_NUMBER

    TRIGGERED_HELLO_PERIOD = 5

    LOGGER = logging.getLogger('hpim.Interface')

    def __init__(self, interface_name: str, vif_index: int):
        self.interface_logger = logging.LoggerAdapter(InterfaceHPIM.LOGGER, {'vif': vif_index,
                                                                             'interfacename': interface_name})

        # Generate BootTime
        self.time_of_boot = int(time.time())

        # Regulate transmission of Hello messages
        self.hello_timer = None

        # protocol neighbors
        self._had_neighbors = False
        self.neighbors = {}
        self.neighbors_lock = RLock()

        # reliable transmission buffer
        self.reliable_transmission_buffer = {} # Key: ID da msg ; value: ReliableMsgTransmission
        self.reliable_transmission_lock = RLock()

        # sequencer for msg reliability
        self.sequencer = 0
        self.sequencer_lock = RLock()

        # security
        self.security_id = 0
        self.security_len = 0
        self.hash_function = None
        self.security_key = b''

        # SOCKET
        ip_interface = netifaces.ifaddresses(interface_name)[netifaces.AF_INET][0]['addr']
        self.ip_interface = ip_interface

        s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_PIM)

        # allow other sockets to bind this port too
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        # explicitly join the multicast group on the interface specified
        #s.setsockopt(socket.SOL_IP, socket.IP_ADD_MEMBERSHIP, socket.inet_aton(Interface.MCAST_GRP) + socket.inet_aton(ip_interface))
        s.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP,
                     socket.inet_aton(Interface.MCAST_GRP) + socket.inet_aton(ip_interface))
        s.setsockopt(socket.SOL_SOCKET, 25, str(interface_name + '\0').encode('utf-8'))

        # set socket output interface
        s.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_IF, socket.inet_aton(ip_interface))

        # set socket TTL to 1
        s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 1)
        s.setsockopt(socket.IPPROTO_IP, socket.IP_TTL, 1)

        # don't receive outgoing packets
        s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, 0)

        super().__init__(interface_name, s, s, vif_index)
        self.force_send_hello()

    @staticmethod
    def _get_address_family():
        """
        Get address family for this interface
        """
        return socket.AF_INET

    @staticmethod
    def get_ip_header_length():
        """"
        Method for getting fixed IPv4 header length. Useful for doing calculations with MTU
        """
        return 20

    def get_ip(self):
        """
        Get IP of this interface
        """
        return self.ip_interface

    @staticmethod
    def get_kernel():
        """
        Get Kernel object
        """
        return Main.kernel

    def _receive(self, raw_bytes, ancdata, src_addr):
        """
        Interface received a new control packet
        """
        if raw_bytes:
            packet = ReceivedPacket(raw_bytes, self)
            if self.is_security_enabled():
                received_security_id = packet.payload.security_id
                if received_security_id != self.security_id:
                    return

                received_security_len = packet.payload.security_length
                if received_security_len != self.security_len:
                    return

                received_security_value = packet.payload.security_value
                received_ip_header = packet.ip_header

                packet.payload.security_value = b''
                calculated_security_value = hmac.new(self.get_security_key(),
                                                     socket.inet_aton(received_ip_header.ip_src) +
                                                     socket.inet_aton(received_ip_header.ip_dst) +
                                                     packet.bytes(), digestmod=self.hash_function).digest()
                if received_security_value != calculated_security_value:
                    return

            self.PKT_FUNCTIONS[packet.payload.get_pim_type()](self, packet)

    def send(self, data: Packet, group_ip: str=MCAST_GRP):
        """
        Send a new packet destined to group_ip IP
        """
        if self.is_security_enabled():
            key = self.get_security_key()
            data.payload.security_id = self.security_id
            data.payload.security_length = self.security_len
            security_value = hmac.new(key, socket.inet_aton(self.get_ip()) + socket.inet_aton(group_ip) +
                                      data.bytes(), digestmod=self.hash_function).digest()
            data.payload.security_value = security_value
        super().send(data=data.bytes(), group_ip=group_ip)

    def is_security_enabled(self):
        return self.get_security_key() != b''

    def get_security_key(self):
        return self.security_key

    def add_security_key(self, security_identifier, security_function, security_key):
        """
        Set Security information for the HMAC of control messages. Set the Security Identifier used to identify
        the Hash algorithm of received control messages, the corresponding hash algorithm and the SecurityKey used to
        calculate the HMAC
        """
        self.security_id = security_identifier
        self.security_len = len(hashlib.new(security_function).digest())
        self.hash_function = getattr(hashlib, security_function)
        self.security_key = str.encode(security_key)

    def remove_security_key(self, security_identifer):
        """
        Disable a given <SecurityIdentifier; HashAlgorithm; SecurityKey> of HMAC control messages.
        """
        if security_identifer != self.security_id:
            return
        self.security_key = b''
        self.security_id = 0
        self.security_len = 0
        self.hash_function = None

    def get_sequence_number(self):
        """
        Get the new sequence number SN to be transmitted in a new control message
        This method will increment the last used sequence number and return it
        It also returns the current BootTime
        """
        with self.sequencer_lock:
            self.sequencer += 1

            if self.sequencer == InterfaceHPIM.MAX_SEQUENCE_NUMBER:
                self.time_of_boot = int(time.time())
                self.sequencer = 1
                self.clear_reliable_transmission()
                self.force_send_hello()

            return (self.time_of_boot, self.sequencer)

    def get_checkpoint_sn(self):
        """
        Get the CheckpointSN to be transmitted in a new Hello message
        """
        print("ENTER CHECK_SN")
        with self.neighbors_lock:
            with self.sequencer_lock:
                with self.reliable_transmission_lock:
                    time_of_boot = self.time_of_boot
                    checkpoint_sn = self.sequencer

                    for rt in self.reliable_transmission_buffer.values():
                        (msg_boot_time, msg_checkpoint_sn) = rt.get_sequence_number()
                        if msg_boot_time == time_of_boot and checkpoint_sn > msg_checkpoint_sn:
                            checkpoint_sn = msg_checkpoint_sn

                    print("EXIT CHECK_SN")
                    return (time_of_boot, checkpoint_sn)

    #Random interval for initial Hello message on bootup or triggered Hello message to a rebooting neighbor
    def force_send_hello(self):
        """
        Force the transmission of a new Hello message
        """
        if self.hello_timer is not None:
            self.hello_timer.cancel()

        hello_timer_time = random.uniform(0, self.TRIGGERED_HELLO_PERIOD)
        self.hello_timer = Timer(hello_timer_time, self.send_hello)
        self.hello_timer.start()

    def send_hello(self):
        """
        Send a new Hello message
        Include in it the HelloHoldTime and CheckpointSN
        """
        self.hello_timer.cancel()

        pim_payload = PacketHPIMHello()
        pim_payload.add_option(PacketHPIMHelloHoldtime(holdtime=4 * HELLO_PERIOD))

        with self.neighbors_lock:
            with self.sequencer_lock:
                with self.reliable_transmission_lock:
                    (bt, checkpoint_sn) = self.get_checkpoint_sn()
                    if bt == self.time_of_boot:
                        pim_payload.add_option(PacketHPIMHelloCheckpointSN(checkpoint_sn))

                ph = PacketHPIMHeader(pim_payload, boot_time=self.time_of_boot)
        packet = Packet(payload=ph)
        self.send(packet)

        # reschedule hello_timer
        self.hello_timer = Timer(HELLO_PERIOD, self.send_hello)
        self.hello_timer.start()

    def remove(self):
        """
        Remove this interface
        Clear all state
        """
        self.hello_timer.cancel()
        self.hello_timer = None

        # send pim_hello timeout message
        pim_payload = PacketHPIMHello()
        pim_payload.add_option(PacketHPIMHelloHoldtime(holdtime=HELLO_HOLD_TIME_TIMEOUT))
        ph = PacketHPIMHeader(pim_payload, boot_time=self.time_of_boot)
        packet = Packet(payload=ph)
        self.send(packet)

        super().remove()
        for n in self.neighbors.values():
            n.remove_neighbor_state()
        self.neighbors.clear()
        self.clear_reliable_transmission()

    def snapshot_multicast_routing_table(self):
        """
        Create a new snapshot
        This method will return the current BootTime, SnapshotSN and all trees to be included in Sync messages
        """
        with self.get_kernel().rwlock.genWlock():
            with self.sequencer_lock:
                (snapshot_bt, snapshot_sn) = self.get_sequence_number()

                trees_to_sync = self.get_kernel().snapshot_multicast_routing_table(self.vif_index)  # type: dict
                tree_to_sync_in_msg_format = {}

                for (source_group, state) in trees_to_sync.items():
                    tree_to_sync_in_msg_format[source_group] = self.create_sync_entry_hdr(source_group[0],
                                                                                          source_group[1],
                                                                                          state.metric_preference,
                                                                                          state.route_metric)
                return (snapshot_bt, snapshot_sn, tree_to_sync_in_msg_format)

    ##############################################
    # Check neighbor status
    ##############################################
    def get_tree_state(self, source_group):
        """
        Get Upstream and Interest state regarding a given tree...
        Regarding Upstream this method will return the Upstream neighbor that offers the best RPC metric
        Regarding Interest this method will return "Interested/True" if at least one neighbor is interested
            or "NotInterested" if all neighbors are not interested in traffic regarding (source_group) tree
        """
        interest_state = False
        upstream_state = None
        for n in list(self.neighbors.values()):
            (neighbor_interest_state, neighbor_upstream_state) = n.get_tree_state(tree=source_group)
            if not interest_state and neighbor_interest_state:
                interest_state = neighbor_interest_state

            if neighbor_upstream_state is not None:
                if upstream_state is None:
                    upstream_state = neighbor_upstream_state
                elif neighbor_upstream_state.is_better_than(upstream_state):
                    upstream_state = neighbor_upstream_state

        print("GET_TREE_STATE_INTERFACE INTEREST:", interest_state)
        print("GET_TREE_STATE_INTERFACE UPSTREAM:", upstream_state)
        return (interest_state, upstream_state)

    def remove_tree_state(self, source_ip, group_ip):
        """
        Remove tree state regarding a given (Source_IP, Group_IP) tree
        """
        for n in list(self.neighbors.values()):
            n.remove_tree_state(source_ip, group_ip)

    # Used to show list of neighbors in CLI interfaces
    def get_neighbors(self):
        """
        Get all neighbors
        """
        return list(self.neighbors.values())

    def get_neighbors_ip(self):
        """
        Get IP of all neighbors
        """
        return set(self.neighbors.keys())

    def force_neighbor_failure(self, neighbor_ip):
        """
        Force an adjacent neighbor to be declared as failed.
        This is used to break an adjacency if a neighbor fails to ack successive control messages
        """
        with self.neighbors_lock:
            neighbor = self.neighbors.get(neighbor_ip, None)
            if neighbor is not None:
                neighbor.remove()

    def did_all_neighbors_acked(self, neighbors_that_acked: set):
        """
        Verify if all neighbor have acknowledged a given message
        Compare if all neighbors that are being monitored are present in neighbors_that_acked set
        """
        return neighbors_that_acked >= self.neighbors.keys()

    def is_neighbor(self, neighbor_ip):
        """
        Verify if neighbor_ip is considered a neighbor
        """
        return neighbor_ip in self.neighbors

    def remove_neighbor(self, ip):
        """
        Remove known neighbor
        """
        with self.neighbors_lock:
            if ip not in self.neighbors:
                return
            self.neighbors.pop(ip)

            # verifiar todas as arvores
            print("REMOVER ANTES RECHECK")
            self.get_kernel().recheck_all_trees(self.vif_index)
            print("REMOVER DEPOIS RECHECK")

    '''
    def change_interface(self):
        # check if ip change was already applied to interface
        old_ip_address = self.ip_interface
        new_ip_interface = netifaces.ifaddresses(self.interface_name)[netifaces.AF_INET][0]['addr']
        if old_ip_address == new_ip_interface:
            return
        
        self._send_socket.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_IF, socket.inet_aton(new_ip_interface))

        self._recv_socket.setsockopt(socket.IPPROTO_IP, socket.IP_DROP_MEMBERSHIP,
                     socket.inet_aton(Interface.MCAST_GRP) + socket.inet_aton(old_ip_address))

        self._recv_socket.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP,
                                     socket.inet_aton(Interface.MCAST_GRP) + socket.inet_aton(new_ip_interface))

        self.ip_interface = new_ip_interface
    '''
    ###########################################
    # Create packets
    ###########################################
    def create_i_am_upstream_msg(self, my_boot_time, sn, source, group, metric_preference, metric):
        ph = PacketHPIMUpstream(source, group, metric_preference, metric, sn)
        return Packet(payload=PacketHPIMHeader(payload=ph, boot_time=my_boot_time))

    def create_i_am_no_longer_upstream_msg(self, my_boot_time, sn, source, group):
        ph = PacketHPIMNoLongerUpstream(source, group, sn)
        return Packet(payload=PacketHPIMHeader(payload=ph, boot_time=my_boot_time))

    def create_interest_msg(self, my_boot_time, sn, source, group):
        ph = PacketHPIMInterest(source, group, sn)
        return Packet(payload=PacketHPIMHeader(payload=ph, boot_time=my_boot_time))

    def create_no_interest_msg(self, my_boot_time, sn, source, group):
        ph = PacketHPIMNoInterest(source, group, sn)
        return Packet(payload=PacketHPIMHeader(payload=ph, boot_time=my_boot_time))

    def create_ack_msg(self, my_boot_time, sn, source, group, neighbor_boot_time, neighbor_snapshot_sn, my_snapshot_sn):
        ack = PacketHPIMAck(source_ip=source, group_ip=group, sequence_number=sn, neighbor_boot_time=neighbor_boot_time,
                            neighbor_snapshot_sn=neighbor_snapshot_sn, my_snapshot_sn=my_snapshot_sn)
        return Packet(payload=PacketHPIMHeader(payload=ack, boot_time=my_boot_time))

    def create_sync_entry_hdr(self, source, group, metric_preference, metric):
        return PacketHPIMSyncEntry(source, group, metric_preference, metric)

    def create_sync_msg(self, my_boot_time, my_snapshot_sn, neighbor_snapshot_sn, sync_sn, upstream_trees,
                        master_flag, more_flag, neighbor_boot_time):
        pkt_sync = PacketHPIMSync(my_snapshot_sn, neighbor_snapshot_sn, sync_sn, upstream_trees,
                                  master_flag, more_flag, neighbor_boot_time)
        return Packet(payload=PacketHPIMHeader(pkt_sync, my_boot_time))

    ###########################################
    # Recv packets
    ###########################################
    def new_neighbor(self, neighbor_ip, boot_time, detected_via_non_sync_msg=True):
        """
        New neighbor detected... start monitoring it and start synchronization
        """
        with self.neighbors_lock:
            self.neighbors[neighbor_ip] = Neighbor(self, neighbor_ip, 120, boot_time, self.time_of_boot)

            if detected_via_non_sync_msg:
                self.neighbors[neighbor_ip].start_sync_process()

    def receive_hello(self, packet):
        """
        Received an Hello packet
        """
        ip = packet.ip_header.ip_src
        boot_time = packet.payload.boot_time
        print("ip = ", ip)
        options = packet.payload.payload.get_options()
        hello_hold_time = options["HOLDTIME"].holdtime
        checkpoint_sn = 0
        if "CHECKPOINT_SN" in options:
            checkpoint_sn = options["CHECKPOINT_SN"].checkpoint_sn
        self.interface_logger.debug('Received Hello message with HelloHoldTime: ' + str(hello_hold_time) +
                                    '; CheckpointSN: ' + str(checkpoint_sn) + ' from neighbor ' + ip)

        with self.neighbors_lock:
            if ip in self.neighbors:
                self.neighbors[ip].recv_hello(boot_time, hello_hold_time, checkpoint_sn)
            else:
                self.new_neighbor(ip, boot_time, True)

    def receive_sync(self, packet):
        """
        Received an Sync packet
        """
        ip = packet.ip_header.ip_src
        boot_time = packet.payload.boot_time

        pkt_hs = packet.payload.payload  # type: PacketHPIMSync

        # Process Sync msg
        my_boot_time = pkt_hs.neighbor_boot_time
        if my_boot_time != self.time_of_boot:
            return

        # All information in Sync msg
        sync_sn = pkt_hs.sync_sequence_number
        upstream_trees = pkt_hs.upstream_trees
        hello_options = pkt_hs.get_hello_options()
        neighbor_sn = pkt_hs.my_snapshot_sn
        my_sn = pkt_hs.neighbor_snapshot_sn
        master_flag = pkt_hs.master_flag
        more_flag = pkt_hs.more_flag

        self.interface_logger.debug('Received Sync message with BootTime: ' + str(boot_time) +
                                    '; NeighborBootTime: ' + str(my_boot_time) +
                                    '; MySnapshotSN: ' + str(neighbor_sn) +
                                    '; NeighborSnapshotSN: ' + str(my_sn) +
                                    '; SyncSN: ' + str(sync_sn) +
                                    '; Master flag: ' + str(master_flag) +
                                    '; More flag: ' + str(more_flag) +
                                    ' from neighbor ' + ip)

        with self.neighbors_lock:
            if ip not in self.neighbors:
                self.new_neighbor(ip, boot_time, detected_via_non_sync_msg=False)

            self.neighbors[ip].recv_sync(upstream_trees, my_sn, neighbor_sn, boot_time, sync_sn, master_flag, more_flag, my_boot_time, hello_options)

    def receive_interest(self, packet):
        """
        Received an Interest packet
        """
        neighbor_source_ip = packet.ip_header.ip_src
        boot_time = packet.payload.boot_time

        pkt_jt = packet.payload.payload  # type: PacketHPIMInterest

        # Process Interest msg
        source_group = (pkt_jt.source, pkt_jt.group)
        sequence_number = pkt_jt.sequence_number

        self.interface_logger.debug('Received Interest message with BootTime: ' + str(boot_time) +
                                    '; Tree: ' + str(source_group) +
                                    '; SN: ' + str(sequence_number) +
                                    ' from neighbor ' + neighbor_source_ip)

        # check neighbor existence
        with self.neighbors_lock:
            neighbor = self.neighbors.get(neighbor_source_ip, None)
            if neighbor is None:
                self.new_neighbor(neighbor_source_ip, boot_time)
                return

            try:
                if neighbor.recv_reliable_packet(sequence_number, source_group, boot_time):
                    if source_group not in neighbor.tree_metric_state:
                        neighbor.tree_interest_state[source_group] = True
                        self.get_kernel().recv_interest_msg(source_group, self)
                    else:
                        neighbor.tree_interest_state[source_group] = True
                        neighbor.tree_metric_state.pop(source_group, None)
                        self.get_kernel().recv_upstream_msg(source_group, self)
            except:
                traceback.print_exc()

    def receive_no_interest(self, packet):
        """
        Received a NoInterest packet
        """
        neighbor_source_ip = packet.ip_header.ip_src
        boot_time = packet.payload.boot_time

        pkt_jt = packet.payload.payload  # type: PacketHPIMNoInterest

        # Process NoInterest msg
        source_group = (pkt_jt.source, pkt_jt.group)
        sequence_number = pkt_jt.sequence_number

        self.interface_logger.debug('Received NoInterest message with BootTime: ' + str(boot_time) +
                                    '; Tree: ' + str(source_group) +
                                    '; SN: ' + str(sequence_number) +
                                    ' from neighbor ' + neighbor_source_ip)

        # check neighbor existence
        with self.neighbors_lock:
            neighbor = self.neighbors.get(neighbor_source_ip, None)
            if neighbor is None:
                self.new_neighbor(neighbor_source_ip, boot_time)
                return

            try:
                if neighbor.recv_reliable_packet(sequence_number, source_group, boot_time):
                    if source_group not in neighbor.tree_metric_state:
                        neighbor.tree_interest_state[source_group] = False
                        self.get_kernel().recv_interest_msg(source_group, self)
                    else:
                        neighbor.tree_interest_state[source_group] = False
                        neighbor.tree_metric_state.pop(source_group, None)
                        self.get_kernel().recv_upstream_msg(source_group, self)
            except:
                traceback.print_exc()

    def receive_i_am_upstream(self, packet):
        """
        Received an IamUpstream packet
        """
        from hpimdm.tree.metric import AssertMetric
        neighbor_source_ip = packet.ip_header.ip_src
        boot_time = packet.payload.boot_time
        pkt_jt = packet.payload.payload # type: PacketHPIMUpstream

        # Process IamUpstream msg
        source_group = (pkt_jt.source, pkt_jt.group)
        sequence_number = pkt_jt.sequence_number

        metric_preference = pkt_jt.metric_preference
        metric = pkt_jt.metric
        received_metric = AssertMetric(metric_preference=metric_preference, route_metric=metric,
                                       ip_address=neighbor_source_ip)

        self.interface_logger.debug('Received IamUpstream message with BootTime: ' + str(boot_time) +
                                    '; Tree: ' + str(source_group) +
                                    '; SN: ' + str(sequence_number) +
                                    '; MetricPreference: ' + str(metric_preference) +
                                    '; Metric: ' + str(metric) +
                                    ' from neighbor ' + neighbor_source_ip)

        # check neighbor existence
        with self.neighbors_lock:
            neighbor = self.neighbors.get(neighbor_source_ip, None)
            if neighbor is None:
                self.new_neighbor(neighbor_source_ip, boot_time)
                return

            try:
                if neighbor.recv_reliable_packet(sequence_number, source_group, boot_time):
                    neighbor.tree_interest_state.pop(source_group, None)
                    neighbor.tree_metric_state[source_group] = received_metric
                    self.get_kernel().recv_upstream_msg(source_group, self)
            except:
                traceback.print_exc()

    def receive_i_am_no_longer_upstream(self, packet):
        """
        Received an IamNoLongerUpstream packet
        """
        neighbor_source_ip = packet.ip_header.ip_src
        boot_time = packet.payload.boot_time
        pkt_jt = packet.payload.payload

        # Process IamNoLongerUpstream msg
        source_group = (pkt_jt.source, pkt_jt.group)
        sequence_number = pkt_jt.sequence_number

        self.interface_logger.debug('Received IamNoLongerUpstream message with BootTime: ' + str(boot_time) +
                                    '; Tree: ' + str(source_group) +
                                    '; SN: ' + str(sequence_number) +
                                    ' from neighbor ' + neighbor_source_ip)

        # check neighbor existence
        with self.neighbors_lock:
            neighbor = self.neighbors.get(neighbor_source_ip, None)
            if neighbor is None:
                self.new_neighbor(neighbor_source_ip, boot_time)
                return

            try:
                if neighbor.recv_reliable_packet(sequence_number, source_group, boot_time):
                    neighbor.tree_interest_state.pop(source_group, None)
                    neighbor.tree_metric_state.pop(source_group, None)
                    self.get_kernel().recv_upstream_msg(source_group, self)
            except:
                traceback.print_exc()

    def receive_ack(self, packet):
        """
        Received an Ack packet
        """
        neighbor_source_ip = packet.ip_header.ip_src
        neighbor_boot_time = packet.payload.boot_time
        pkt_ack = packet.payload.payload  # type: PacketHPIMAck

        # Process Ack msg
        source_group = (pkt_ack.source, pkt_ack.group)
        my_boot_time = pkt_ack.neighbor_boot_time
        my_snapshot_sn = pkt_ack.neighbor_snapshot_sn
        neighbor_snapshot_sn = pkt_ack.my_snapshot_sn
        sequence_number = pkt_ack.sequence_number

        self.interface_logger.debug('Received Ack message with BootTime: ' + str(neighbor_boot_time) +
                                    '; NeighborBootTime: ' + str(my_boot_time) +
                                    '; MySnapshotSN: ' + str(neighbor_snapshot_sn) +
                                    '; NeighborSnapshotSN: ' + str(my_snapshot_sn) +
                                    '; Tree: ' + str(source_group) +
                                    '; SN: ' + str(sequence_number) +
                                    ' from neighbor ' + neighbor_source_ip)

        # check neighbor existence
        with self.neighbors_lock:
            neighbor = self.neighbors.get(neighbor_source_ip, None) # type: Neighbor
            if neighbor is None:
                self.new_neighbor(neighbor_source_ip, neighbor_boot_time)
                return

            with self.sequencer_lock:
                with self.reliable_transmission_lock:
                    if not neighbor.recv_ack(my_boot_time, neighbor_boot_time, my_snapshot_sn, neighbor_snapshot_sn):
                        return

                    #if my_boot_time != self.time_of_boot:
                    #    return

                    reliable_transmission = self.reliable_transmission_buffer.get(source_group, None)
                    if reliable_transmission is not None:
                        reliable_transmission.receive_ack(neighbor_source_ip, my_boot_time, sequence_number)


    PKT_FUNCTIONS = {
        PacketHPIMHello.PIM_TYPE:            receive_hello,
        PacketHPIMSync.PIM_TYPE:             receive_sync,
        PacketHPIMUpstream.PIM_TYPE:         receive_i_am_upstream,
        PacketHPIMNoLongerUpstream.PIM_TYPE: receive_i_am_no_longer_upstream,
        PacketHPIMInterest.PIM_TYPE:         receive_interest,
        PacketHPIMNoInterest.PIM_TYPE:       receive_no_interest,
        PacketHPIMAck.PIM_TYPE:              receive_ack,
    }


    ########################################################################
    # Message Transmission
    ########################################################################
    def get_reliable_message_transmission(self, tree):
        """
        Get object used to monitor the reliable transmission of messages regarding a given tree
        """
        with self.reliable_transmission_lock:
            reliable_msg_transmission = self.reliable_transmission_buffer.get(tree, None)

            if reliable_msg_transmission is None:
                reliable_msg_transmission = ReliableMessageTransmission(self)
                self.reliable_transmission_buffer[tree] = reliable_msg_transmission

            return reliable_msg_transmission

    def send_i_am_upstream(self, source, group, rpc):
        """
        Send a new IamUpstream message
        """
        tree = (source, group)
        with self.sequencer_lock:
            with self.reliable_transmission_lock:
                self.get_reliable_message_transmission(tree).send_i_am_upstream(source, group, rpc)

    def send_i_am_no_longer_upstream(self, source, group):
        """
        Send a new IamNoLongerUpstream message
        """
        tree = (source, group)
        with self.sequencer_lock:
            with self.reliable_transmission_lock:
                self.get_reliable_message_transmission(tree).send_i_am_no_longer_upstream(source, group)

    def send_interest(self, source, group, dst):
        """
        Send a new Interest message
        """
        tree = (source, group)
        with self.sequencer_lock:
            with self.reliable_transmission_lock:
                self.get_reliable_message_transmission(tree).send_interest(source, group, dst)

    def send_no_interest(self, source, group, dst):
        """
        Send a new NoInterest message
        """
        tree = (source, group)
        with self.sequencer_lock:
            with self.reliable_transmission_lock:
                self.get_reliable_message_transmission(tree).send_no_interest(source, group, dst)

    def neighbor_start_synchronization(self, neighbor_ip, my_snapshot_bt, my_snapshot_sn):
        """
        Neighbor started a new synchronization... consider all trees included in the snapshot to be Acknowledged
        (all packets transmitted with a lower BootTime or lower SN compared to SnapshotSN will be considered to have been
        acknowledged by it)
        """
        with self.reliable_transmission_lock:
            for rmt in self.reliable_transmission_buffer.values():
                rmt.receive_ack(neighbor_ip, my_snapshot_bt, my_snapshot_sn)

    def cancel_all_messages(self, tree):
        """
        Cancel the reliable monitoring of all messages regarding a given tree
        """
        with self.reliable_transmission_lock:
            if tree in self.reliable_transmission_buffer:
                self.reliable_transmission_buffer[tree].cancel_all_messages()

    def cancel_interest_message(self, tree, neighbor_ip):
        """
        Cancel the reliable monitoring of all interest messages regarding a given tree
        """
        with self.reliable_transmission_lock:
            if tree in self.reliable_transmission_buffer:
                self.reliable_transmission_buffer[tree].cancel_message_unicast(neighbor_ip)

    def cancel_upstream_message(self, tree):
        """
        Cancel the reliable monitoring of all upstream messages regarding a given tree
        """
        with self.reliable_transmission_lock:
            if tree in self.reliable_transmission_buffer:
                self.reliable_transmission_buffer[tree].cancel_message_multicast()

    def clear_reliable_transmission(self):
        """
        Cancel the reliable monitoring of all messages regarding all trees
        """
        with self.reliable_transmission_lock:
            for rmt in self.reliable_transmission_buffer.values():
                rmt.cancel_all_messages()

    '''
    def cancel_message(self, tree):
        with self.reliable_transmission_lock:
            reliable_packet = self.reliable_transmission_buffer.get(tree, None)
            if reliable_packet is not None:
                reliable_packet.cancel_message()
    '''



#!/usr/bin/env python3

import os
import sys
import socket
import argparse
import traceback
import faulthandler
import _pickle as pickle
from hpimdm.tree import hpim_globals
from hpimdm.daemon.Daemon import Daemon
from hpimdm import Main

VERSION = "1.3.3.2"


def client_socket(data_to_send):
    """
    Send command to daemon process through a socket
    """
    # Create a UDS socket
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

    # Connect the socket to the port where the server is listening
    server_address = '/tmp/hpim_uds_socket' + str(hpim_globals.MULTICAST_TABLE_ID)
    #print('connecting to %s' % server_address)
    try:
        sock.connect(server_address)
        sock.sendall(pickle.dumps(data_to_send))
        data_rcv = sock.recv(1024 * 256)
        if data_rcv:
            print(pickle.loads(data_rcv))
    except socket.error:
        pass
    finally:
        #print('closing socket')
        sock.close()


class MyDaemon(Daemon):
    def run(self):
        """
        Daemon process will run this method until the daemon process explicitly is stopped
        """
        Main.main()
        server_address = '/tmp/hpim_uds_socket' + str(hpim_globals.MULTICAST_TABLE_ID)

        # Make sure the socket does not already exist
        try:
            os.unlink(server_address)
        except OSError:
            if os.path.exists(server_address):
                raise

        # Create a UDS socket
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        # Bind the socket to the port
        sock.bind(server_address)

        # Listen for incoming connections
        sock.listen(1)
        while True:
            try:
                connection, client_address = sock.accept()
                data = connection.recv(256 * 1024)
                print(sys.stderr, 'sending data back to the client')
                print(pickle.loads(data))
                args = pickle.loads(data)
                if 'ipv4' not in args or 'ipv6' not in args or not (args.ipv4 or args.ipv6):
                    args.ipv4 = True
                    args.ipv6 = False

                if 'list_interfaces' in args and args.list_interfaces:
                    connection.sendall(pickle.dumps(Main.list_enabled_interfaces(ipv4=args.ipv4, ipv6=args.ipv6)))
                elif 'list_neighbors' in args and args.list_neighbors:
                    connection.sendall(pickle.dumps(Main.list_neighbors(ipv4=args.ipv4, ipv6=args.ipv6)))
                elif 'list_state' in args and args.list_state:
                    connection.sendall(pickle.dumps(Main.list_state(ipv4=args.ipv4, ipv6=args.ipv6)))
                elif 'list_neighbors_state' in args and args.list_neighbors_state:
                    connection.sendall(pickle.dumps(Main.list_neighbors_state(ipv4=args.ipv4, ipv6=args.ipv6)))
                elif 'list_sequence_numbers' in args and args.list_sequence_numbers:
                    connection.sendall(pickle.dumps(Main.list_sequence_numbers(ipv4=args.ipv4, ipv6=args.ipv6)))
                elif 'add_interface' in args and args.add_interface:
                    Main.add_hpim_interface(args.add_interface[0], ipv4=args.ipv4, ipv6=args.ipv6)
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'flood_initial_data' in args and args.flood_initial_data:
                    connection.sendall(pickle.dumps(Main.change_initial_flood_setting()))
                elif 'hold_forwarding_state' in args and args.hold_forwarding_state:
                    connection.sendall(pickle.dumps(Main.hold_forwarding_state()))
                elif 'add_interface_igmp' in args and args.add_interface_igmp:
                    Main.add_membership_interface(interface_name=args.add_interface_igmp[0], ipv4=True, ipv6=False)
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'add_interface_mld' in args and args.add_interface_mld:
                    Main.add_membership_interface(interface_name=args.add_interface_mld[0], ipv4=False, ipv6=True)
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'remove_interface' in args and args.remove_interface:
                    Main.remove_interface(args.remove_interface[0], hpim=True, ipv4=args.ipv4, ipv6=args.ipv6)
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'remove_interface_igmp' in args and args.remove_interface_igmp:
                    Main.remove_interface(args.remove_interface_igmp[0], membership=True, ipv4=True, ipv6=False)
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'remove_interface_mld' in args and args.remove_interface_mld:
                    Main.remove_interface(args.remove_interface_mld[0], membership=True, ipv4=False, ipv6=True)
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'list_hmac_algorithms' in args and args.list_hmac_algorithms:
                    connection.sendall(pickle.dumps(Main.list_hash_algorithms()))
                elif 'add_interface_security' in args and args.add_interface_security:
                    Main.add_security_key(args.add_interface_security[0], int(args.add_interface_security[1]),
                                          args.add_interface_security[2], args.add_interface_security[3],
                                          ipv4=args.ipv4, ipv6=args.ipv6)
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'remove_interface_security' in args and args.remove_interface_security:
                    Main.remove_security_key(args.remove_interface_security[0], int(args.remove_interface_security[1]),
                                             ipv4=args.ipv4, ipv6=args.ipv6)
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'stop' in args and args.stop:
                    Main.stop()
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'test' in args and args.test:
                    Main.test(args.test[0], args.test[1])
                    connection.shutdown(socket.SHUT_RDWR)
                elif 'traceback' in args and args.traceback:
                    faulthandler.dump_traceback(file=sys.stderr, all_threads=True)
                    connection.shutdown(socket.SHUT_RDWR)
            except Exception:
                connection.shutdown(socket.SHUT_RDWR)
                traceback.print_exc()
            finally:
                # Clean up the connection
                connection.close()


def main():
    """
    Entry point for HPIM-DM
    """
    parser = argparse.ArgumentParser(description='HPIM-DM protocol', prog='hpim-dm')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-start", "--start", action="store_true", default=False,
                       help="Start Protocol")
    group.add_argument("-stop", "--stop", action="store_true", default=False,
                       help="Stop Protocol")
    group.add_argument("-restart", "--restart", action="store_true", default=False,
                       help="Restart Protocol")
    group.add_argument("-li", "--list_interfaces", action="store_true", default=False,
                       help="List All Interfaces. Use -4 or -6 to specify IPv4 or IPv6 interface.")
    group.add_argument("-ln", "--list_neighbors", action="store_true", default=False,
                       help="List All Neighbors. Use -4 or -6 to specify IPv4 or IPv6 HPIM neighbors.")
    group.add_argument("-ls", "--list_state", action="store_true", default=False,
                       help="List state of IGMP/MLD and HPIM-DM. "
                            "Use -4 or -6 to specify IPv4 or IPv6 state respectively.")
    group.add_argument("-lns", "--list_neighbors_state", action="store_true", default=False,
                       help="List Upstream and Interest state of all neighbors. "
                            "Use -4 or -6 to specify IPv4 or IPv6 HPIM neighbor state.")
    group.add_argument("-lsn", "--list_sequence_numbers", action="store_true", default=False,
                       help="List Sequence Numbers. Use -4 or -6 to list stored IPv4 or IPv6 HPIM sequence numbers.")
    group.add_argument("-mr", "--multicast_routes", action="store_true", default=False,
                       help="List Multicast Routing table. "
                            "Use -4 or -6 to specify IPv4 or IPv6 multicast routing table.")
    group.add_argument("-fid", "--flood_initial_data", action="store_true", default=False,
                       help="Flood initial data packets")
    group.add_argument("-hfs", "--hold_forwarding_state", action="store_true", default=False,
                       help="Hold forwarding state during a small amount of time after AW interface becomes AL " +
                            "(prevent loss of data packets after AW replacement)")
    group.add_argument("-ai", "--add_interface", nargs=1, metavar='INTERFACE_NAME',
                       help="Add HPIM-DM interface. Use -4 or -6 to specify IPv4 or IPv6 interface.")
    group.add_argument("-aiigmp", "--add_interface_igmp", nargs=1, metavar='INTERFACE_NAME',
                       help="Add IGMP interface")
    group.add_argument("-aimld", "--add_interface_mld", nargs=1, metavar='INTERFACE_NAME',
                       help="Add MLD interface")
    group.add_argument("-ri", "--remove_interface", nargs=1, metavar='INTERFACE_NAME',
                       help="Remove HPIM-DM interface. Use -4 or -6 to specify IPv4 or IPv6 interface.")
    group.add_argument("-riigmp", "--remove_interface_igmp", nargs=1, metavar='INTERFACE_NAME',
                       help="Remove IGMP interface")
    group.add_argument("-rimld", "--remove_interface_mld", nargs=1, metavar='INTERFACE_NAME',
                       help="Remove MLD interface")
    group.add_argument("-lsec", "--list_hmac_algorithms", action="store_true", default=False,
                       help="List available HMAC Hash algorithms")
    group.add_argument("-aisec", "--add_interface_security", nargs=4, metavar=('INTERFACE_NAME', "SECURITY_IDENTIFIER",
                                                                               "HASH_FUNCTION", "KEY"),
                       help="Add security information to interface INTERFACE_NAME. Control messages will be secured " +
                            " with SECURITY_IDENTIFIER, HMAC algorithm based on HASH_FUNCTION and key KEY. " +
                            "Use -4 or -6 to specify IPv4 or IPv6 HPIM interface. " +
                            "To determine available hash functions run -lsec command")
    group.add_argument("-risec", "--remove_interface_security", nargs=2, metavar=('INTERFACE_NAME',
                                                                                  "SECURITY_IDENTIFIER"),
                       help="Remove security information identified by SECURITY_IDENTIFIER from interface INTERFACE_NAME."
                            " Use -4 or -6 to specify IPv4 or IPv6 interface.")
    group.add_argument("-v", "--verbose", action="store_true", default=False,
                       help="Verbose (print all debug messages)")
    group.add_argument("-t", "--test", nargs=2, metavar=('ROUTER_NAME', 'SERVER_LOG_IP'),
                       help="Tester... send log information to SERVER_LOG_IP. Set the router name to ROUTER_NAME")
    group.add_argument("-traceback", "--traceback", action="store_true", default=False,
                       help="Dump the tracebacks of all threads into file")
    group.add_argument("--version", action='version', version='%(prog)s ' + VERSION)
    group_ipversion = parser.add_mutually_exclusive_group(required=False)
    group_ipversion.add_argument("-4", "--ipv4", action="store_true", default=False, help="Setting for IPv4")
    group_ipversion.add_argument("-6", "--ipv6", action="store_true", default=False, help="Setting for IPv6")
    args = parser.parse_args()

    #print(parser.parse_args())
    # This script must be run as root!
    if os.geteuid() != 0:
        sys.exit('HPIM-DM must be run as root!')

    daemon = MyDaemon('/tmp/Daemon-hpim' + str(hpim_globals.MULTICAST_TABLE_ID) + '.pid')
    if args.start:
        print("start")
        daemon.start()
        sys.exit(0)
    elif args.stop:
        client_socket(args)
        daemon.stop()
        sys.exit(0)
    elif args.restart:
        daemon.restart()
        sys.exit(0)
    elif args.verbose:
        os.system("tail -f /var/log/hpimdm/stdout" + str(hpim_globals.MULTICAST_TABLE_ID))
        sys.exit(0)
    elif args.multicast_routes:
        if args.ipv4 or not args.ipv6:
            os.system("ip mroute show table " + str(hpim_globals.MULTICAST_TABLE_ID))
        elif args.ipv6:
            os.system("ip -6 mroute show table " + str(hpim_globals.MULTICAST_TABLE_ID))
        sys.exit(0)
    elif not daemon.is_running():
        print("HPIM-DM is not running")
        parser.print_usage()
        sys.exit(0)

    client_socket(args)


if __name__ == "__main__":
    main()



import json
import struct
from .PacketHPIMHello import PacketHPIMHelloJson, PacketHPIMHello
from .PacketHPIMIamUpstream import PacketHPIMUpstreamJson, PacketHPIMUpstream, PacketHPIMUpstream_v6
from .PacketHPIMNotUpstream import PacketHPIMNoLongerUpstreamJson, PacketHPIMNoLongerUpstream,\
    PacketHPIMNoLongerUpstream_v6
from .PacketHPIMInterest import PacketHPIMInterestJson, PacketHPIMNoInterestJson, PacketHPIMInterest,\
    PacketHPIMNoInterest, PacketHPIMInterest_v6, PacketHPIMNoInterest_v6
from .PacketHPIMAck import PacketHPIMAckJson, PacketHPIMAck, PacketHPIMAck_v6
from .PacketHPIMSync import PacketHPIMSyncJson, PacketHPIMSync, PacketHPIMSync_v6

from .PacketPayload import PacketPayload
'''
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|  Type  |   msg........                                        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''
class PacketHPIMHeaderJson(PacketPayload):

    PIM_MSG_TYPES = {"HELLO": PacketHPIMHelloJson,
                     "INTEREST": PacketHPIMInterestJson,
                     "NO_INTEREST": PacketHPIMNoInterestJson,
                     "I_AM_UPSTREAM": PacketHPIMUpstreamJson,
                     "I_AM_NO_LONGER_UPSTREAM": PacketHPIMNoLongerUpstreamJson,
                     "ACK": PacketHPIMAckJson,
                     "SYNC": PacketHPIMSyncJson,
                    }

    def __init__(self, payload, boot_time=0):
        self.payload = payload
        self.boot_time = boot_time

    def get_pim_type(self):
        return self.payload.PIM_TYPE

    def bytes(self) -> bytes:
        """
        Obtain Protocol Packet in a format to be transmitted (JSON)
        This method will return the Header and Payload in JSON format
        """
        # obter mensagem e criar checksum
        data = {"TYPE": self.get_pim_type(),
                "BOOT_TIME": self.boot_time,
                "DATA": self.payload.bytes()}
        msg = json.dumps(data).encode()
        return msg

    def __len__(self):
        return len(self.bytes())

    @staticmethod
    def parse_bytes(data: bytes):
        """
        Parse received Packet from JSON and convert it into Header object... also parse Header's payload
        """
        msg = json.loads(data.decode())

        pkt_type = msg["TYPE"]
        print("TYPE", pkt_type)

        id_reliable = msg["BOOT_TIME"]

        pim_payload = msg["DATA"]
        print("DATA", pim_payload)
        pim_payload = PacketHPIMHeaderJson.PIM_MSG_TYPES[pkt_type].parse_bytes(pim_payload)
        return PacketHPIMHeaderJson(pim_payload, id_reliable)



'''
HEADER IN BYTE FORMAT
 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                            BootTime                           |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|Version| Type  |      Security Identifier      |Security Length|
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                         Security Value                        |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
'''

class PacketHPIMHeader(PacketPayload):
    PIM_VERSION = 0

    PIM_HDR = "! L B H B"
    PIM_HDR_LEN = struct.calcsize(PIM_HDR)

    PIM_MSG_TYPES = {PacketHPIMHello.PIM_TYPE: PacketHPIMHello,
                     PacketHPIMSync.PIM_TYPE: PacketHPIMSync,
                     PacketHPIMUpstream.PIM_TYPE: PacketHPIMUpstream,
                     PacketHPIMNoLongerUpstream.PIM_TYPE: PacketHPIMNoLongerUpstream,
                     PacketHPIMInterest.PIM_TYPE: PacketHPIMInterest,
                     PacketHPIMNoInterest.PIM_TYPE: PacketHPIMNoInterest,
                     PacketHPIMAck.PIM_TYPE: PacketHPIMAck,
                     }

    def __init__(self, payload, boot_time=0, security_id=0, security_length=0, security_value=b''):
        self.payload = payload
        self.boot_time = boot_time
        self.security_id = security_id
        self.security_length = security_length
        self.security_value = security_value

    @property
    def security_value(self):
        if self.security_length == 0:
            return b''
        else:
            len_diff = abs(len(self._security_value) - self.security_length)
            return self._security_value + b'\x00' * len_diff

    @security_value.setter
    def security_value(self, security_value):
        self._security_value = security_value

    def get_pim_type(self):
        return self.payload.PIM_TYPE

    def bytes(self) -> bytes:
        """
        Obtain Protocol Packet in a format to be transmitted (binary)
        This method will return the Header and Payload in binary format
        """
        pim_vrs_type = (PacketHPIMHeader.PIM_VERSION << 4) + self.get_pim_type()
        msg = struct.pack(PacketHPIMHeader.PIM_HDR, self.boot_time, pim_vrs_type, self.security_id,
                          self.security_length)
        msg += self.security_value + self.payload.bytes()
        return msg

    def __len__(self):
        return len(self.bytes())

    @classmethod
    def parse_bytes(cls, data: bytes):
        """
        Parse received Packet from bits/bytes and convert them into Header object... also parse Header's payload
        """
        print("parsePimHdr: ", data)

        pim_hdr = data[0:cls.PIM_HDR_LEN]
        (boot_time, pim_ver_type, security_id, security_len) = struct.unpack(cls.PIM_HDR, pim_hdr)

        print(pim_ver_type)
        pim_version = (pim_ver_type & 0xF0) >> 4
        pim_type = pim_ver_type & 0x0F

        if pim_version != cls.PIM_VERSION:
            print("Version of PROTOCOL packet received not known (!=0)")
            raise Exception

        security_and_pim_payload = data[cls.PIM_HDR_LEN:]
        security_value = security_and_pim_payload[:security_len]
        print("Received hmac value: ", security_value)
        pim_payload = security_and_pim_payload[security_len:]
        pim_payload = cls.PIM_MSG_TYPES[pim_type].parse_bytes(pim_payload)
        return PacketHPIMHeader(pim_payload, boot_time, security_id, security_len, security_value)


class PacketHPIMHeader_v6(PacketHPIMHeader):
    PIM_MSG_TYPES = {PacketHPIMHello.PIM_TYPE: PacketHPIMHello,
                     PacketHPIMSync.PIM_TYPE: PacketHPIMSync_v6,
                     PacketHPIMUpstream.PIM_TYPE: PacketHPIMUpstream_v6,
                     PacketHPIMNoLongerUpstream.PIM_TYPE: PacketHPIMNoLongerUpstream_v6,
                     PacketHPIMInterest.PIM_TYPE: PacketHPIMInterest_v6,
                     PacketHPIMNoInterest.PIM_TYPE: PacketHPIMNoInterest_v6,
                     PacketHPIMAck.PIM_TYPE: PacketHPIMAck_v6,
                     }

    def __init__(self, payload, boot_time=0, security_id=0, security_length=0, security_value=b''):
        super().__init__(payload, boot_time, security_id, security_length, security_value)



# MLD timers (in seconds)
ROBUSTNESS_VARIABLE = 2
QUERY_INTERVAL = 125
QUERY_RESPONSE_INTERVAL = 10
MULTICAST_LISTENER_INTERVAL = (ROBUSTNESS_VARIABLE * QUERY_INTERVAL) + (QUERY_RESPONSE_INTERVAL)
OTHER_QUERIER_PRESENT_INTERVAL = (ROBUSTNESS_VARIABLE * QUERY_INTERVAL) + 0.5 * QUERY_RESPONSE_INTERVAL
STARTUP_QUERY_INTERVAL = (1 / 4) * QUERY_INTERVAL
STARTUP_QUERY_COUNT = ROBUSTNESS_VARIABLE
LAST_LISTENER_QUERY_INTERVAL = 1
LAST_LISTENER_QUERY_COUNT = ROBUSTNESS_VARIABLE
UNSOLICITED_REPORT_INTERVAL = 10


# MLD msg type
MULTICAST_LISTENER_QUERY_TYPE = 130
MULTICAST_LISTENER_REPORT_TYPE = 131
MULTICAST_LISTENER_DONE_TYPE = 132


from hpimdm.utils import TYPE_CHECKING

if TYPE_CHECKING:
    from .tree_if_root import TreeInterfaceRoot

'''
class SFMRRootState:
    @staticmethod
    def tree_transitions_to_active_state(interface: 'TreeInterfaceRoot') -> None:
        """
        Tree transitions to Active state AND
        interface roles dont change
        """
        interface.logger.debug('tree_transitions_to_active_state')
        interface.send_my_interest()

    @staticmethod
    def tree_remains_in_active_state_and_non_root_transitions_to_root_interface(interface: 'TreeInterfaceRoot'):
        """
        Tree remains in Active state AND
        Non-root interface transitions to Root type
        """
        interface.logger.debug('tree_remains_in_active_state_and_non_root_transitions_to_root_interface')
        interface.send_i_am_no_longer_upstream()
        interface.send_my_interest()

    @staticmethod
    def tree_transitions_from_active_to_inactive_state_due_to_transition_from_non_root_to_root_interface(interface: 'TreeInterfaceRoot'):
        """
        Tree transitions from Active to Inactive state AND
        Non-Root interface transitions to Root type
        """
        interface.logger.debug('tree_transitions_from_active_to_inactive_state_due_to_transition_from_non_root_to_root_interface')
        interface.send_i_am_no_longer_upstream()

    @staticmethod
    def tree_transitions_to_active_state_and_non_root_interface_transitions_to_root(interface: 'TreeInterfaceRoot') -> None:
        """
        Tree transitions to Active state AND
        Non-Root interface transitions to Root type
        """
        interface.logger.debug('tree_transitions_to_active_state_and_non_root_interface_transitions_to_root')
        SFMRRootState.tree_transitions_to_active_state(interface)

    @staticmethod
    def transition_to_it_or_ot_and_active_tree(interface: 'TreeInterfaceRoot') -> None:
        """
        Tree is in Active state AND
        interface becomes interested or not interested in receiving data packets
        """
        interface.logger.debug('transition_to_it_or_ot_and_active_tree')
        interface.send_my_interest()

    @staticmethod
    def tree_is_active_and_best_upstream_router_reelected(interface: 'TreeInterfaceRoot') -> None:
        """
        Tree is in Active state AND
        BestUpstream neighbor changes or BestUpstream neighbor sends an IamUpstream message and remains responsible for
        forwarding data packets
        """
        interface.logger.debug('tree_is_active_and_best_upstream_router_reelected')
        interface.send_my_interest()
'''

class SFMRNewRootState:
    @staticmethod
    def interfaces_roles_change_and_tree_remains_active(interface: 'TreeInterfaceRoot'):
        """
        Interfaces roles change (NonRoot->Root) AND
        Tree was Active and remains Active
        """
        interface.logger.debug('interfaces_roles_change_and_tree_remains_active')
        interface.send_i_am_no_longer_upstream()
        interface.send_my_interest()

    @staticmethod
    def interfaces_roles_change_and_tree_was_unsure_and_transitions_to_active(interface: 'TreeInterfaceRoot'):
        """
        Interfaces roles change (NonRoot->Root) AND
        Tree was Unsure and transitions to Active state
        """
        interface.logger.debug('interfaces_roles_change_and_tree_was_unsure_and_transitions_to_active')
        interface.send_my_interest()

    @staticmethod
    def interfaces_roles_change_and_tree_was_active_and_transitions_to_unsure_and_best_upstream_neighbor_is_null(interface: 'TreeInterfaceRoot'):
        """
        Interfaces roles change (NonRoot->Root) AND
        Tree transitions from Active to Unsure state AND
        BestUpstreamNeighbor is null
        """
        interface.logger.debug('interfaces_roles_change_and_tree_was_active_and_transitions_to_unsure_and_best_upstream_neighbor_is_null')
        interface.send_i_am_no_longer_upstream()

    @staticmethod
    def interfaces_roles_change_and_tree_was_active_and_transitions_to_unsure_and_best_upstream_neighbor_not_null(interface: 'TreeInterfaceRoot'):
        """
        Interfaces roles change (NonRoot->Root) AND
        Tree transitions from Active to Unsure state AND
        BestUpstreamNeighbor is not null (possible loop detected)
        """
        interface.logger.debug('interfaces_roles_change_and_tree_was_active_and_transitions_to_unsure_and_best_upstream_neighbor_not_null')
        interface.send_i_am_no_longer_upstream()
        interface.send_my_interest()

    @staticmethod
    def interfaces_roles_change_and_tree_remains_unsure_and_best_upstream_neighbor_not_null(interface: 'TreeInterfaceRoot'):
        """
        Interfaces roles change (NonRoot->Root) AND
        Tree was Unsure and remains Unsure AND
        BestUpstreamNeighbor is not null (possible loop detected)
        """
        interface.logger.debug('interfaces_roles_change_and_tree_remains_unsure_and_best_upstream_neighbor_not_null')
        interface.send_my_interest()

    @staticmethod
    def interfaces_roles_dont_change_and_router_transition_to_it_or_ot(interface: 'TreeInterfaceRoot'):
        """
        Interfaces roles dont change (this interface remains Root) AND
        BestUpstreamNeighbor does not change AND
        router changes its interest in receiving data packets (becomes interested or not interested)
        """
        interface.logger.debug('interfaces_roles_dont_change_and_router_transition_to_it_or_ot')
        interface.send_my_interest()


    @staticmethod
    def interfaces_roles_dont_change_and_best_upstream_neighbor_reelected(interface: 'TreeInterfaceRoot'):
        """
        Interfaces roles dont change (this interface remains Root) AND
        BestUpstreamNeighbor changes AND
        BestUpstreamNeighbor is not null
        (other router was elected as the responsible for forwarding multicast data packets)

        OR

        Interfaces roles dont change (this interface remains Root) AND
        BestUpstreamNeighbor transmitted a new IamUpstream message AND
        that neighbor is still responsible for forwarding multicast data packet
        (a new IamUpstream message was received from the BestUpstreamNeighbor but that router is still responsible for
        forwarding multicast data packets)
        """
        interface.logger.debug('interfaces_roles_dont_change_and_best_upstream_neighbor_reelected')
        interface.send_my_interest()
